module OSFileSelect(osSelectInputFile, osSelectOutputFile, osSelectDirectory) where


--	Clean Object I/O library, version 1.2


import	ClCrossCall_12
import	OSEvent
import	CommonDef (dumpFatalError)
import	Cutil_12(int2addr, addr2int)
import  CString(newCString, peekCString)
import  MarshalAlloc(free)


osFileSelectFatalError :: String -> String -> x
osFileSelectFatalError function error
	= dumpFatalError function "OSFileSelect" error

osSelectInputFile :: (OSEvent->s->IO s) -> s -> IO (Maybe String,s)
osSelectInputFile handleOSEvent state = do
	(rcci,state) <- issueCleanRequest (callback handleOSEvent) (rq0Cci ccRqFILEOPENDIALOG) state
	mb_name <- getInputFileName rcci
	return (mb_name,state)
	where
		getInputFileName :: CrossCallInfo -> IO (Maybe String)
		getInputFileName cci
			| ccMsg cci == ccRETURN2 =
				if p1 cci == 0
				then return Nothing
				else do
					let pathnamePtr = (int2addr (p2 cci))
					pathname <- peekCString pathnamePtr
					free pathnamePtr
					return (Just pathname)
			| ccMsg cci == ccWASQUIT =
				return Nothing
			| otherwise =
				osFileSelectFatalError "osSelectInputFile" ("unexpected ccMsg field of return CrossCallInfo ("++show (ccMsg cci)++")")

osSelectOutputFile :: (OSEvent->s->IO s) -> s -> String -> String -> IO (Maybe String,s)
osSelectOutputFile handleOSEvent state prompt filename = do
	promptPtr <- newCString prompt
	filenamePtr <- newCString filename
	(rcci,state) <- issueCleanRequest (callback handleOSEvent) (rq2Cci ccRqFILESAVEDIALOG (addr2int promptPtr) (addr2int filenamePtr)) state
	free promptPtr
	free filenamePtr
	mb_name	<- getOutputFileName rcci
	return (mb_name,state)
	where
		getOutputFileName :: CrossCallInfo -> IO (Maybe String)
		getOutputFileName cci
			| ccMsg cci == ccRETURN2 =
				if p1 cci == 0
				then return Nothing
				else do
					let pathPtr = (int2addr (p2 cci))
					path <- peekCString pathPtr
					free pathPtr
					return (Just path)
			| ccMsg cci == ccWASQUIT =
				return Nothing
			| otherwise =
				osFileSelectFatalError "osSelectOutputFile" ("unexpected ccMsg field of return CrossCallInfo ("++show (ccMsg cci)++")")

osSelectDirectory :: (OSEvent->s->IO s) -> s -> IO (Maybe String,s)
osSelectDirectory handleOSEvent state = do
	(rcci,state) <- issueCleanRequest (callback handleOSEvent) (rq0Cci ccRqDIRECTORYDIALOG) state
	mb_name	<- getInputFileName rcci
	return (mb_name,state)
	where
		getInputFileName :: CrossCallInfo -> IO (Maybe String)
		getInputFileName cci
			| ccMsg cci == ccRETURN2 =
				if p1 cci == 0
				then return Nothing
				else do
					let pathnamePtr = (int2addr (p2 cci))
					pathname <- peekCString pathnamePtr
					free pathnamePtr
					return (Just pathname)
			| ccMsg cci == ccWASQUIT =
				return Nothing
			| otherwise =
				osFileSelectFatalError "osSelectDirectory" ("unexpected ccMsg field of return CrossCallInfo ("++show (ccMsg cci)++")")

--	callback lifts a function::(OSEvent -> s -> s) to
--        a crosscallfunction::(CrossCallInfo -> s -> IO (CrossCallInfo,s))
callback :: (OSEvent->s->IO s) -> CrossCallInfo -> s -> IO (CrossCallInfo,s)
callback handleOSEvent cci state = handleOSEvent cci state >>= \s -> return (return0Cci,s)
