module ControlAccess ( getcontrolslayouts, getcontrolsviewsizes
					 , getcontrolsoutersizes, getcontrolsselects
					 , getcontrolsshowstates, getcontrolstexts ) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	ControlAccess contains all read functions on controls.
--	********************************************************************************


import Id
import StdIOCommon
import CommonDef
import WindowAccess
import StdControlAttribute
import OSWindow(OSWindowMetrics, osGetEditControlText, osScrollbarsAreVisible)


{-	Higher order access on [WElementHandle ls ps].
	Although statemapWElementHandles has (... WElementHandle ls ps ...) function arguments,
	these will only be applied to WItemHandle alternatives.
-}

type MapFunction ps x = forall ls . WElementHandle ls ps -> x -> IO (WElementHandle ls ps, x)

statemapWElementHandles :: MapFunction ps x -> [Id] -> [WElementHandle ls ps] -> x -> IO ([Id], [WElementHandle ls ps], x)
statemapWElementHandles f ids [] s = return (ids, [], s)
statemapWElementHandles f ids (itemH:itemHs) s
	| null ids      = return (ids, (itemH:itemHs), s)
	| otherwise   = do
		(ids, itemH,  s) <- statemapWElementHandle  f ids itemH  s
		(ids, itemHs, s) <- statemapWElementHandles f ids itemHs s
		return (ids, itemH:itemHs, s)
	where
		statemapWElementHandle :: MapFunction ps x -> [Id] -> WElementHandle ls ps -> x -> IO ([Id], WElementHandle ls ps, x)
		statemapWElementHandle f ids (WListLSHandle itemHs) s = do
			(ids, itemHs, s) <- statemapWElementHandles f ids itemHs s
			return (ids, WListLSHandle itemHs, s)
		statemapWElementHandle f ids (WChangeLSHandle chLS itemHs) s = do
			(ids, itemHs, s) <- statemapWElementHandles f ids itemHs s
			return (ids, WChangeLSHandle chLS itemHs, s)
		statemapWElementHandle f ids (WExtendLSHandle exLS itemHs) s = do
			(ids, itemHs, s) <- statemapWElementHandles f ids itemHs s
			return (ids, WExtendLSHandle exLS itemHs, s)
		statemapWElementHandle f ids itemH s = do
			(ids, itemH, s) <- statemapWElementHandle' f ids itemH s
			(if isRecursiveControl (wItemKind itemH)
			 then do
				(ids, itemHs, s) <- statemapWElementHandles f ids (wItems itemH) s
				return (ids, itemH{wItems=itemHs}, s)
			 else return (ids, itemH, s))
			where
				statemapWElementHandle' :: MapFunction ps x -> [Id] -> WElementHandle ls ps -> x -> IO ([Id], WElementHandle ls ps, x)
				statemapWElementHandle' f ids itemH@(WItemHandle {wItemId=wItemId}) s
					| isNothing wItemId = return (ids, itemH, s)
					| not hadId			= return (ids, itemH, s)
					| otherwise			= do
						(itemH, s) <- f itemH s
						return (ids1, itemH, s)
					where				
						itemId			= fromJust wItemId
						(hadId,ids1)	= removeCheck itemId ids


--	Access operations on WElementHandle.

getcontrolslayouts :: [Id] -> WIDS -> WindowHandle ls ps -> [(Id,Bool,(Maybe ItemPos,Vector2))] -> IO (WindowHandle ls ps, [(Id,Bool,(Maybe ItemPos,Vector2))])
getcontrolslayouts ids wids wH@(WindowHandle {whItems=itemHs}) layouts = do
	(ids, itemHs, layouts) <- statemapWElementHandles getlayouts ids itemHs layouts
	return (wH{whItems=itemHs}, layouts)
	where
		getlayouts :: WElementHandle ls ps -> [(Id,Bool,(Maybe ItemPos,Vector2))] -> IO (WElementHandle ls ps, [(Id,Bool,(Maybe ItemPos,Vector2))])
		getlayouts itemH@(WItemHandle {wItemAtts=atts,wItemPos=wItemPos}) layouts =
			return (itemH, layouts1)
			where
				itemPos			= if hasAtt then Just (getControlPosAtt posAtt) else Nothing
				(hasAtt,posAtt)	= cselect isControlPos (ControlPos (StdIOCommon.Left,zero)) atts
				layout			= (itemId,True,(itemPos,toVector wItemPos))
				(_,layouts1)	= creplace (eqfst3id itemId) layout layouts
				itemId          = fromJust (wItemId itemH)

getcontrolsviewsizes :: [Id] -> OSWindowMetrics -> WIDS -> WindowHandle ls ps -> [(Id,Bool,Size)] -> IO (WindowHandle ls ps , [(Id,Bool,Size)])
getcontrolsviewsizes ids wMetrics wids wH@(WindowHandle {whItems=itemHs}) sizes = do
		(ids, itemHs, sizes) <- statemapWElementHandles (getsizes wMetrics) ids itemHs sizes
		return (wH{whItems=itemHs}, sizes)
		where
			getsizes :: OSWindowMetrics -> WElementHandle ls ps -> [(Id,Bool,Size)] -> IO (WElementHandle ls ps, [(Id,Bool,Size)])
			getsizes wMetrics itemH@(WItemHandle {wItemKind=wItemKind}) sizes =
				return (itemH, sizes1)
				where
					itemSize		= wItemSize itemH
					info			= getWItemCompoundInfo (wItemInfo itemH)
					(domainRect,hasScrolls)
									= (compoundDomain info,(isJust (compoundHScroll info),isJust (compoundVScroll info)))
					visScrolls		= osScrollbarsAreVisible wMetrics domainRect (toTuple itemSize) hasScrolls
					size			= if (wItemKind/=IsCompoundControl)
 									  then itemSize
									  else rectSize (getCompoundContentRect wMetrics visScrolls (sizeToRect itemSize))
					(_,sizes1)		= creplace (eqfst3id itemId) (itemId,True,size) sizes
					itemId          = fromJust (wItemId itemH)

getcontrolsoutersizes :: [Id] -> OSWindowMetrics -> WIDS -> WindowHandle ls ps -> [(Id,Bool,Size)] -> IO (WindowHandle ls ps , [(Id,Bool,Size)])
getcontrolsoutersizes ids wMetrics wids wH@(WindowHandle {whItems=itemHs}) sizes = do
	(ids, itemHs, sizes) <- statemapWElementHandles (getsizes wMetrics) ids itemHs sizes
	return (wH{whItems=itemHs}, sizes)
	where
		getsizes :: OSWindowMetrics -> WElementHandle ls ps -> [(Id,Bool,Size)] -> IO (WElementHandle ls ps, [(Id,Bool,Size)])
		getsizes wMetrics itemH@(WItemHandle {}) sizes =
			return (itemH, sizes1)
			where
				size			= wItemSize itemH
				(_,sizes1)		= creplace (eqfst3id itemId) (itemId,True,size) sizes
				itemId          = fromJust (wItemId itemH)

getcontrolsselects :: [Id] -> OSWindowMetrics -> WIDS -> WindowHandle ls ps -> [(Id,Bool,SelectState)] -> IO (WindowHandle ls ps , [(Id,Bool,SelectState)])
getcontrolsselects ids wMetrics wids wH@(WindowHandle {whItems=itemHs}) selects = do
	(ids, itemHs, selects) <- statemapWElementHandles getselects ids itemHs selects
	return (wH{whItems=itemHs}, selects)
	where
		getselects :: WElementHandle ls ps -> [(Id,Bool,SelectState)] -> IO (WElementHandle ls ps, [(Id,Bool,SelectState)])
		getselects itemH@(WItemHandle {wItemSelect=wItemSelect}) selects =
			return (itemH, selects1)
			where
				selectstate		= if wItemSelect then Able else Unable
				select			= (itemId,True,selectstate)
				(_,selects1)	= creplace (eqfst3id itemId) select selects
				itemId          = fromJust (wItemId itemH)
				

getcontrolsshowstates :: [Id] -> OSWindowMetrics -> WIDS -> WindowHandle ls ps -> [(Id,Bool,Bool)] -> IO (WindowHandle ls ps, [(Id,Bool,Bool)])
getcontrolsshowstates ids wMetrics wids wH@(WindowHandle {whItems=itemHs}) shows = do
	(ids, itemHs, shows) <- statemapWElementHandles getshowstates ids itemHs shows
	return (wH{whItems=itemHs}, shows)
	where
		getshowstates :: WElementHandle ls ps -> [(Id,Bool,Bool)] -> IO (WElementHandle ls ps, [(Id,Bool,Bool)])
		getshowstates itemH@(WItemHandle {wItemShow=wItemShow}) shows =
			return (itemH, shows1)
			where
				show			= (itemId,True,wItemShow)
				(_,shows1)		= creplace (eqfst3id itemId) show shows
				itemId          = fromJust (wItemId itemH)

getcontrolstexts :: [Id] -> OSWindowMetrics -> WIDS -> WindowHandle ls ps -> [(Id,Bool,Maybe String)] -> IO (WindowHandle ls ps, [(Id,Bool,Maybe String)])
getcontrolstexts ids wMetrics wids wH@(WindowHandle {whItems=itemHs}) texts = do
	(ids, itemHs, texts) <- statemapWElementHandles gettext ids itemHs texts
	return (wH{whItems=itemHs}, texts)
	where
		gettext :: WElementHandle ls ps -> [(Id,Bool,Maybe String)] -> IO (WElementHandle ls ps, [(Id,Bool,Maybe String)])
		gettext itemH@(WItemHandle {}) texts =			
			case wItemKind itemH of
					IsPopUpControl  -> do
							(textline, popUpInfo) <- getPopUpText (getWItemPopUpInfo  info)
							return (itemH{wItemInfo=WPopUpInfo popUpInfo}, snd (creplace (eqfst3id itemId) (itemId,True,Just textline) texts))
					IsTextControl  -> do
							let textline = textInfoText   (getWItemTextInfo   info)
							return (itemH, snd (creplace (eqfst3id itemId) (itemId,True,Just textline) texts))
					IsEditControl  -> do
							textline <- osGetEditControlText (wPtr wids) (wItemPtr itemH)
							return (itemH{wItemInfo=WEditInfo ((getWItemEditInfo info){editInfoText=textline})}, snd (creplace (eqfst3id itemId) (itemId,True,Just textline) texts))
					IsButtonControl -> do
							let textline = buttonInfoText (getWItemButtonInfo info)
							return (itemH, snd (creplace (eqfst3id itemId) (itemId,True,Just textline) texts))
					_ -> return (itemH, texts)
			where
				info                 = wItemInfo itemH
				itemId               = fromJust (wItemId itemH)
				
				getPopUpText :: PopUpInfo ls ps -> IO (String, PopUpInfo ls ps)
				getPopUpText popUpInfo@(PopUpInfo {popUpInfoEdit=popUpInfoEdit,popUpInfoItems=popUpInfoItems}) =
					case popUpInfoEdit of
						Nothing -> 
							if null popUpInfoItems
							then return ("", popUpInfo)
							else return (fst (popUpInfoItems!!(max 0 (popUpInfoIndex popUpInfo - 1))), popUpInfo)
						Just info -> do
							content <- osGetEditControlText (wItemPtr itemH) (popUpEditPtr info)
							return (content, popUpInfo{popUpInfoEdit=Just (info{popUpEditText=content})})

