{-# OPTIONS -cpp #-}

module ProcessDevice (processFunctions) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	ProcessDevice defines the window device event handlers.
--	********************************************************************************


import CommonDef
import IOState
import OSDocumentInterface
import ProcessEvent
import StdProcessAttribute
import Toolbar
#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


processdeviceFatalError :: String -> String -> x
processdeviceFatalError rule error
	= dumpFatalError rule "ProcessDevice" error


processFunctions :: DeviceFunctions IF_MVAR(,ps)
processFunctions
	= DeviceFunctions
		{ dDevice = ProcessDevice
		, dShow	  = return
		, dHide	  = return
		, dEvent  = processEvent
		, dDoIO   = processIO
		, dOpen   = processOpen
		, dClose  = processClose
		}

processOpen :: IF_MVAR(GUI (),ps -> GUI ps ps)
processOpen IF_MVAR(,ps)
	= do {
		hasProcess <- accIOEnv (ioStHasDevice ProcessDevice);
		if   hasProcess
		then return IF_MVAR((),ps)
		else 
		do {
			appIOEnv (ioStSetDeviceFunctions processFunctions);
			osdinfo <- accIOEnv ioStGetOSDInfo;
			createOSDInfo osdinfo;
			return IF_MVAR((),ps)
		   }
	     }
	where
		createOSDInfo :: OSDInfo -> GUI IF_MVAR(,ps) ()
		createOSDInfo emptyOSDInfo
			| di==NDI = appIOEnv (ioStSetOSDInfo emptyOSDInfo)
			| di==MDI = do
				atts <- accIOEnv ioStGetProcessAttributes
				let acceptOpenFiles	= any isProcessOpenFiles atts
				let hasToolbarAtt	= any isProcessToolbar   atts
				osdinfo <- liftIO (osOpenMDI (not hasToolbarAtt) acceptOpenFiles)
				appIOEnv (ioStSetOSDInfo osdinfo)
				openToolbar
			| di==SDI = do
				atts <- accIOEnv ioStGetProcessAttributes
				let acceptOpenFiles = any isProcessOpenFiles atts					
				osdinfo <- liftIO (osOpenSDI acceptOpenFiles)
				appIOEnv (ioStSetOSDInfo osdinfo)
				openToolbar
			where
				di = getOSDInfoDocumentInterface emptyOSDInfo

processClose :: IF_MVAR(GUI (),ps -> GUI ps ps)
processClose IF_MVAR(,ps)
	= do {
		accIOEnv (ioStGetDevice ProcessDevice);
		appIOEnv (ioStRemoveDeviceFunctions ProcessDevice);
		return IF_MVAR((),ps)
	  }

processIO :: DeviceEvent -> IF_MVAR(,ps ->) GUI IF_MVAR((),ps ps)

{-	ProcessInitialise is the first event sent to an interactive process.
	This allows the system to evaluate its initialisation action. No further
	actions are required.
-}
processIO ProcessInitialise IF_MVAR(,ps)
	= return IF_MVAR((),ps)

processIO ProcessRequestClose IF_MVAR(,ps)
	= do {
		atts                 <- accIOEnv ioStGetProcessAttributes;
		let (hasCloseAtt,att) = cselect isProcessClose undefined atts
		in
		if   not hasCloseAtt
		then return IF_MVAR((),ps)
		else getProcessCloseFun att IF_MVAR(,ps)
	  }
processIO (ProcessRequestOpenFiles openFilesInfo) ps = do
	atts                 <- accIOEnv ioStGetProcessAttributes;
	let (hasFilesOpenAtt,att) = cselect isProcessOpenFiles undefined atts
	(if not hasFilesOpenAtt
	 then return ps
	 else getProcessOpenFilesFun att openFilesInfo ps)

processIO _ IF_MVAR(,_)
	= processdeviceFatalError "processIO" "unexpected DeviceEvent"
