{-# OPTIONS -cpp #-}

module StdControlAttribute (module StdControlAttribute, module StdControlDef) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	StdControlAttribute specifies which ControlAttributes are valid for each of the
--	standard controls.
--	********************************************************************************


import StdControlDef
import StdPictureDef(PenAttribute(..))
import StdPicture(Look)

#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


isValidButtonControlAttribute :: ControlAttribute IF_MVAR(,ls ps) -> Bool
isValidButtonControlAttribute (ControlFunction     _) = True
isValidButtonControlAttribute (ControlHide	    ) = True
isValidButtonControlAttribute (ControlId           _) = True
isValidButtonControlAttribute (ControlModsFunction _) = True
isValidButtonControlAttribute (ControlPos          _) = True
isValidButtonControlAttribute (ControlSelectState  _) = True
isValidButtonControlAttribute (ControlWidth        _) = True
isValidButtonControlAttribute _                       = False

isValidCheckControlAttribute :: ControlAttribute IF_MVAR(,ls ps) -> Bool
isValidCheckControlAttribute (ControlHide	  ) = True
isValidCheckControlAttribute (ControlId          _) = True
isValidCheckControlAttribute (ControlPos         _) = True
isValidCheckControlAttribute (ControlSelectState _) = True
isValidCheckControlAttribute (ControlTip         _) = True
isValidCheckControlAttribute _			    = False

isValidCompoundControlAttribute :: (ControlAttribute IF_MVAR(,ls ps)) -> Bool
isValidCompoundControlAttribute (ControlFunction     _)	= False
isValidCompoundControlAttribute (ControlModsFunction _)	= False
isValidCompoundControlAttribute _						= True

isValidEditControlAttribute :: ControlAttribute IF_MVAR(,ls ps) -> Bool
isValidEditControlAttribute (ControlActivate    _)  = True
isValidEditControlAttribute (ControlDeactivate  _)  = True
isValidEditControlAttribute (ControlHide	 )  = True
isValidEditControlAttribute (ControlId          _)  = True
isValidEditControlAttribute (ControlKeyboard _ _ _) = True
isValidEditControlAttribute (ControlPos         _)  = True
isValidEditControlAttribute (ControlResize      _)  = True
isValidEditControlAttribute (ControlSelectState _)  = True
isValidEditControlAttribute (ControlTip         _)  = True
isValidEditControlAttribute _                       = False

isValidLayoutControlAttribute :: ControlAttribute IF_MVAR(,ls ps) -> Bool
isValidLayoutControlAttribute (ControlHide	   ) = True
isValidLayoutControlAttribute (ControlHMargin   _ _) = True
isValidLayoutControlAttribute (ControlId          _) = True
isValidLayoutControlAttribute (ControlItemSpace _ _) = True
isValidLayoutControlAttribute (ControlMinimumSize _) = True
isValidLayoutControlAttribute (ControlOuterSize   _) = True
isValidLayoutControlAttribute (ControlPos         _) = True
isValidLayoutControlAttribute (ControlResize      _) = True
isValidLayoutControlAttribute (ControlSelectState _) = True
isValidLayoutControlAttribute (ControlViewSize    _) = True
isValidLayoutControlAttribute (ControlVMargin   _ _) = True
isValidLayoutControlAttribute _			     = False

isValidPopUpControlAttribute :: ControlAttribute IF_MVAR(,ls ps) -> Bool
isValidPopUpControlAttribute (ControlActivate    _) = True
isValidPopUpControlAttribute (ControlDeactivate  _) = True
isValidPopUpControlAttribute (ControlHide	  ) = True
isValidPopUpControlAttribute (ControlId          _) = True
isValidPopUpControlAttribute (ControlPos         _) = True
isValidPopUpControlAttribute (ControlSelectState _) = True
isValidPopUpControlAttribute (ControlTip         _) = True
isValidPopUpControlAttribute (ControlWidth	 _) = True
isValidPopUpControlAttribute _			    = False

isValidRadioControlAttribute :: ControlAttribute IF_MVAR(,ls ps) -> Bool
isValidRadioControlAttribute (ControlHide	  ) = True
isValidRadioControlAttribute (ControlId          _) = True
isValidRadioControlAttribute (ControlPos         _) = True
isValidRadioControlAttribute (ControlSelectState _) = True
isValidRadioControlAttribute (ControlTip         _) = True
isValidRadioControlAttribute _			    = False

isValidSliderControlAttribute :: ControlAttribute IF_MVAR(,ls ps) -> Bool
isValidSliderControlAttribute (ControlHide	   ) = True
isValidSliderControlAttribute (ControlId          _) = True
isValidSliderControlAttribute (ControlPos         _) = True
isValidSliderControlAttribute (ControlResize      _) = True
isValidSliderControlAttribute (ControlSelectState _) = True
isValidSliderControlAttribute (ControlTip         _) = True
isValidSliderControlAttribute _			     = False

isValidTextControlAttribute :: ControlAttribute IF_MVAR(,ls ps) -> Bool
isValidTextControlAttribute (ControlId    _) = True
isValidTextControlAttribute (ControlPos   _) = True
isValidTextControlAttribute (ControlWidth _) = True
isValidTextControlAttribute _                = False


isControlActivate 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlActivate 	(ControlActivate _) 	= True
isControlActivate 	_			= False

isControlDeactivate 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlDeactivate 	(ControlDeactivate _) = True
isControlDeactivate 	_			  = False


isControlFunction 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlFunction 	(ControlFunction _)   = True
isControlFunction 	_                     = False

isControlHide 		:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlHide 		ControlHide		= True
isControlHide 		_				= False

isControlHMargin 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlHMargin 	(ControlHMargin _ _)	= True
isControlHMargin 	_			= False

isControlHScroll 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlHScroll 	(ControlHScroll _)	= True
isControlHScroll 	_			= False

isControlId 		:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlId 		(ControlId _)         	= True
isControlId 		_                     	= False

isControlItemSpace 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlItemSpace 	(ControlItemSpace _ _) 	= True
isControlItemSpace 	_			= False

isControlKeyboard 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlKeyboard 	(ControlKeyboard _ _ _) = True
isControlKeyboard 	_                       = False

isControlLook		:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlLook		(ControlLook _ _)	= True
isControlLook		_			= False

isControlMinimumSize	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlMinimumSize	(ControlMinimumSize _)	= True
isControlMinimumSize	_			= False

isControlModsFunction 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlModsFunction 	(ControlModsFunction _) = True
isControlModsFunction 	_                       = False

isControlMouse 		:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlMouse 		(ControlMouse _ _ _)	= True
isControlMouse 		_			= False

isControlOrigin 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlOrigin 	(ControlOrigin _)	= True
isControlOrigin		_			= False

isControlOuterSize 	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlOuterSize 	(ControlOuterSize _)	= True
isControlOuterSize 	_			= False

isControlPen 		:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlPen 		(ControlPen _)		= True
isControlPen 		_			= False

isControlPos 		:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlPos 		(ControlPos _)   	= True
isControlPos 		_                	= False

isControlResize		:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlResize		(ControlResize _)	= True
isControlResize		_			= False

isControlSelectState	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlSelectState	(ControlSelectState _)	= True
isControlSelectState	_						= False

isControlTip		:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlTip		(ControlTip _)		= True
isControlTip		_			= False

isControlViewDomain	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlViewDomain	(ControlViewDomain _)	= True
isControlViewDomain	_			= False

isControlViewSize	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlViewSize	(ControlViewSize _)	= True
isControlViewSize	_			= False

isControlVMargin	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlVMargin	(ControlVMargin _ _)	= True
isControlVMargin	_			= False

isControlVScroll	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlVScroll	(ControlVScroll _)	= True
isControlVScroll	_			= False

isControlWidth    	:: ControlAttribute IF_MVAR(,ls ps) -> Bool
isControlWidth    	(ControlWidth _)        = True
isControlWidth    	_                       = False

getControlActivateFun :: ControlAttribute IF_MVAR(,ls ps) -> IF_MVAR(GUIFun (),GUIFun ls ps)
getControlActivateFun (ControlActivate f) = f

getControlDeactivateFun :: ControlAttribute IF_MVAR(,ls ps) -> IF_MVAR(GUIFun (),GUIFun ls ps)
getControlDeactivateFun (ControlDeactivate f) = f

getControlFun :: ControlAttribute IF_MVAR(,ls ps) -> IF_MVAR(GUIFun (),GUIFun ls ps)
getControlFun (ControlFunction f) = f

getControlHMarginAtt :: ControlAttribute IF_MVAR(,ls ps) -> (Int,Int)
getControlHMarginAtt (ControlHMargin left right) = (left,right)

getControlHScrollFun :: ControlAttribute IF_MVAR(,ls ps) -> ScrollFunction
getControlHScrollFun (ControlHScroll f) = f

getControlIdAtt :: ControlAttribute IF_MVAR(,ls ps) -> Id
getControlIdAtt (ControlId id) = id

getControlItemSpaceAtt :: ControlAttribute IF_MVAR(,ls ps) -> (Int,Int)
getControlItemSpaceAtt (ControlItemSpace hspace vspace) = (hspace,vspace)

getControlKeyboardAtt :: ControlAttribute IF_MVAR(,ls ps) -> (KeyboardStateFilter,SelectState,KeyboardFunction IF_MVAR(,ls ps))
getControlKeyboardAtt (ControlKeyboard filter s f) = (filter,s,f)

getControlLookAtt :: ControlAttribute IF_MVAR(,ls ps) -> (Bool,Look)
getControlLookAtt (ControlLook systemLook f) = (systemLook,f)

getControlMinimumSizeAtt :: ControlAttribute IF_MVAR(,ls ps) -> Size
getControlMinimumSizeAtt (ControlMinimumSize size) = size

getControlModsFun :: ControlAttribute IF_MVAR(,ls ps) -> ModifiersFunction IF_MVAR(,ls ps)
getControlModsFun (ControlModsFunction f) = f

getControlMouseAtt :: ControlAttribute IF_MVAR(,ls ps) -> (MouseStateFilter,SelectState,MouseFunction IF_MVAR(,ls ps))
getControlMouseAtt (ControlMouse filter s f) = (filter,s,f)

getControlOriginAtt :: ControlAttribute IF_MVAR(,ls ps) -> Point2
getControlOriginAtt (ControlOrigin p) = p

getControlOuterSizeAtt :: ControlAttribute IF_MVAR(,ls ps) -> Size
getControlOuterSizeAtt (ControlOuterSize size) = size

getControlPenAtt :: ControlAttribute IF_MVAR(,ls ps) -> [PenAttribute]
getControlPenAtt (ControlPen atts) = atts

getControlPosAtt :: ControlAttribute IF_MVAR(,ls ps) -> ItemPos
getControlPosAtt (ControlPos itemPos) = itemPos

getControlResizeFun :: ControlAttribute IF_MVAR(,ls ps) -> ControlResizeFunction
getControlResizeFun (ControlResize f) = f

getControlSelectStateAtt :: ControlAttribute IF_MVAR(,ls ps) -> SelectState
getControlSelectStateAtt (ControlSelectState s) = s

getControlTipAtt :: ControlAttribute IF_MVAR(,ls ps) -> String
getControlTipAtt (ControlTip tip) = tip

getControlViewDomainAtt :: ControlAttribute IF_MVAR(,ls ps) -> ViewDomain
getControlViewDomainAtt (ControlViewDomain pd) = pd

getControlViewSizeAtt :: ControlAttribute IF_MVAR(,ls ps) -> Size
getControlViewSizeAtt (ControlViewSize size) = size

getControlVMarginAtt :: ControlAttribute IF_MVAR(,ls ps) -> (Int,Int)
getControlVMarginAtt (ControlVMargin top bottom) = (top,bottom)

getControlVScrollFun :: ControlAttribute IF_MVAR(,ls ps) -> ScrollFunction
getControlVScrollFun (ControlVScroll f) = f

getControlWidthAtt :: ControlAttribute IF_MVAR(,ls ps) -> ControlWidth
getControlWidthAtt (ControlWidth w) = w

