{-# OPTIONS -cpp #-}

module StdGUI (
#if MVAR
#else
		noLS, noLS1, 
#endif
		GUIFun, GUI, ProcessAttribute(..), ModifiersFunction, KeyboardFunction,
		MouseFunction, SliderAction, ScrollFunction,
		ProcessWindowResizeFunction, ProcessOpenFilesFunction, ToolbarItem(..)) where

--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	This is a new module that provides similar functionality as StdPSt in the Clean
--	Object I/O library.
--	********************************************************************************


import {-# SOURCE #-} IOState(GUI)
import StdIOCommon(Modifiers(..),KeyboardState(..),SliderState(..),SliderMove(..),MouseState(..),ViewFrame, ItemPos, Size(..))
import StdBitmap(Bitmap)
#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


#if MVAR
#else
{-	Lift an action (ps -> GUI ps ps) to a GUIFun ls ps. -}
noLS :: (ps -> GUI ps ps) -> GUIFun ls ps	-- Lift action GUI ps to GUIFun ls ps
noLS action (ls,ps) = do { ps1 <- action ps; return (ls,ps1) }

{-	Lift an action (ps -> GUI ps ps) that requires x to a GUIFun ls ps that requires x. -}
noLS1 :: (x -> ps -> GUI ps ps) -> x -> GUIFun ls ps
noLS1 action x (ls,ps) = do { ps1 <- action x ps; return (ls,ps1) }
#endif

#if MVAR
type	GUIFun a
 =	GUI a
#else
type	GUIFun ls ps
 =	(ls,ps) -> GUI ps (ls,ps)
#endif


{-	Process attributes.				-}

data	ProcessAttribute ps					 -- Default:
	= ProcessActivate	(ps -> GUI ps ps)		 -- No action on activate
	| ProcessDeactivate	(ps -> GUI ps ps)		 -- No action on deactivate
	| ProcessClose 		(ps -> GUI ps ps)		 -- Process is closed
--	Attributes for (M/S)DI process only:
	| ProcessOpenFiles	(ProcessOpenFilesFunction ps) 	 -- Request to open files
	| ProcessWindowPos	ItemPos				 -- Platform dependent
	| ProcessWindowSize	Size				 -- Platform dependent
	| ProcessWindowResize	(ProcessWindowResizeFunction ps) -- Platform dependent
 	| ProcessToolbar	[ToolbarItem ps]	 	 -- Process has no toolbar
 --	Attributes for MDI processes only:
	| ProcessNoWindowMenu					 -- Process has WindowMenu

type	ProcessWindowResizeFunction ps 
	 =	Size						-- Old ProcessWindow size
	 ->	Size						-- New ProcessWindow size
	 ->	ps -> GUI ps ps
type	ProcessOpenFilesFunction ps
	 =	[String]					-- The file names to open
	 -> ps -> GUI ps ps

data	ToolbarItem ps
	= ToolbarItem Bitmap (Maybe String) (ps -> GUI ps ps)
	| ToolbarSeparator



{-	Frequently used function types.			-}

#if MVAR
type	ModifiersFunction = Modifiers     -> GUIFun ()
type	KeyboardFunction  = KeyboardState -> GUIFun ()
type	MouseFunction	  = MouseState	  -> GUIFun ()
type	SliderAction 	  = SliderMove 	  -> GUIFun ()
#else
type	ModifiersFunction ls ps = Modifiers     -> GUIFun ls ps
type	KeyboardFunction  ls ps = KeyboardState -> GUIFun ls ps
type	MouseFunction	  ls ps = MouseState	-> GUIFun ls ps
type	SliderAction 	  ls ps = SliderMove 	-> GUIFun ls ps
#endif


{-	Scrolling function.				-}

type	ScrollFunction =
		ViewFrame	->				-- Current	view
		SliderState	->				-- Current	state of scrollbar
		SliderMove	->				-- Action of the user
		Int			
