{-# OPTIONS -cpp #-}

module StdId ( Ids(..)
	     , openIds, openRIds
             , getParentId
             , Id.Id
             , IOState.GUI, IOState.IOSt
             ) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	StdId specifies the generation functions for identification values.
--	********************************************************************************


import Concurrent
import Id
import IOState
import Maybe
import World
#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


class Monad envM => Ids envM where
	openId   ::        envM Id
	openRId  ::        envM (RId m r)

instance Ids IO where
	openId  = do
		i <- loadWorld
		storeWorld (i-1)
		return (toId i)
	
	openRId = do
		i   <- loadWorld
		cIn  <- newChan
		cOut <- newChan
		storeWorld (i-1)
		return (toRId i cIn cOut)

instance Ids (GUI IF_MVAR(,ps)) where
	openId  = do
		i <- ioStGetIdSeed
		ioStSetIdSeed (i-1)
		return (toId i)
	
	openRId = do
		i   <- ioStGetIdSeed;
		cIn  <- liftIO newChan;
		cOut <- liftIO newChan;
		ioStSetIdSeed (i-1);
		return (toRId i cIn cOut)


openIds  :: Ids envM => Int -> envM [Id]
openIds n = sequence (replicate n openId)

openRIds :: Ids envM => Int -> envM [RId m r]
openRIds n = sequence (replicate n openRId)

getParentId :: Id -> GUI IF_MVAR(,ps) (Maybe Id)
getParentId id
	= ioStGetIdTable >>= (\idtable -> return (fmap idpId (getIdParent id idtable)))
