{-# OPTIONS -cpp #-}

module StdWindow ( Windows(..), Dialogs(..), module StdWindowDef
                 , closeWindow, closeActiveWindow, getActiveWindow
                 , getWindowLook, setWindowLook
                 , enableWindowMouse, disableWindowMouse
                 , getWindowViewFrame, moveWindowViewFrame
                 , drawInWindow
                 , setWindowViewDomain, getWindowViewDomain
                 , setWindowScrollFunction, getWindowScrollFunction
                 , setWindowTitle, setWindowCursor
                 , getWindowTitle, getWindowOk
                 , getWindowCancel, getWindowCursor
                 ) where


--	********************************************************************************
--	Clean to Haskell Standard Object I/O library, version 1.2
--	
--	StdWindow defines functions on windows.
--	********************************************************************************

import List(find)
import Monad(when)
import CommonDef
import ControlValidate
import ControlLayout(layoutControls)
import ControlRelayout(relayoutControls)
import qualified ControlPos as CP
import Id
import IOState
import StdControlClass
import StdWindowDef
import StdWindowAttribute
import WindowAccess
import WindowCreate
import WindowDevice
import WindowDispose
import WindowValidate
import WindowClipState(validateWindowClipState, forceValidWindowClipState)
import qualified WindowDraw as WD
import WindowDraw(drawWindowLook)
import OSPicture(Draw(..))
import WindowUpdate(updateWindowBackgrounds)
import OSWindow
#include "preprocess.h"
         -- Compile with -cpp; set MVAR=1 for MVAR version


stdWindowFatalError :: String -> String -> x
stdWindowFatalError function error
	= dumpFatalError function "StdWindow" error

class Windows wdef where
	openWindow :: ls -> wdef ls ps -> ps -> GUI ps ps

	
--	Functions applied to non-existent windows or unknown ids have no effect.
class Dialogs ddef where
#if MVAR
	openDialog :: ddef -> GUI ()
	openModalDialog	:: ddef -> GUI ()
#else
	openDialog :: ls -> ddef ls ps -> ps -> GUI ps ps
	openModalDialog	:: ls -> ddef ls ps -> ps -> GUI ps (Maybe ls)
#endif

 
instance Controls c => Windows (Window c) where	
	openWindow IF_MVAR(,ls) (Window title controls atts) IF_MVAR(,ps)	
		= do {
			IF_MVAR(,ps1 <-) dOpen windowFunctions IF_MVAR(,ps);
			isZero <- accIOEnv checkZeroWindowBound;
			if isZero then throwGUI ErrorViolateDI
			else do {
			   maybe_okId <- validateWindowId maybe_id;
			   case maybe_okId of
			   	Nothing -> throwGUI ErrorIdsInUse
			   	Just okId -> do
				  	cs   <- controlToHandles controls
				  	it   <- ioStGetIdTable
				  	rt   <- ioStGetReceiverTable
				  	ioId <- accIOEnv ioStGetIOId
				  	let itemHs = map controlStateToWElementHandle cs
				  	(case controlIdsAreConsistent ioId okId itemHs rt it of
				  		Nothing -> throwGUI ErrorIdsInUse
				  		Just (it,rt1) -> 
				  			let it1 = snd (addIdToIdTable okId (IdParent {idpIOId=ioId,idpDevice=WindowDevice,idpId=okId}) it)
				  			in do
				  				ioStSetIdTable it1
				  				ioStSetReceiverTable rt1
								ps2 <- openwindow okId (WindowLSHandle {wlsState=ls,wlsHandle=initWindowHandle title Modeless IsWindow NoWindowInfo itemHs atts}) ps1
								appIOEnv decreaseWindowBound
								return ps2)
				  }
			  }
		where
			maybe_id = getWindowIdAttribute atts


instance (Controls c) => Dialogs (Dialog c) where
	openDialog IF_MVAR(,ls) (Dialog title controls atts) IF_MVAR(,ps)
		= do {
			IF_MVAR(,ps1 <-) dOpen windowFunctions IF_MVAR(,ps);
			maybe_okId <- validateWindowId maybe_id;
			case maybe_okId of
				Nothing   -> throwGUI ErrorIdsInUse
				Just okId -> do
					cs   <- controlToHandles controls
					it   <- ioStGetIdTable
					rt   <- ioStGetReceiverTable
					ioId <- accIOEnv ioStGetIOId
					let itemHs          = map controlStateToWElementHandle cs				    
				    	(case controlIdsAreConsistent ioId okId itemHs rt it of
				    		Nothing -> throwGUI ErrorIdsInUse
				    		Just (it,rt1) ->
				    			let it1 = snd (addIdToIdTable okId (IdParent {idpIOId=ioId,idpDevice=WindowDevice,idpId=okId}) it)
				    			in do
				    				ioStSetIdTable it1
				    				ioStSetReceiverTable rt1
								openwindow okId (WindowLSHandle {wlsState=ls,wlsHandle=initWindowHandle title Modeless IsDialog NoWindowInfo itemHs atts}) ps1)
		}
		where
			maybe_id = getWindowIdAttribute atts


	openModalDialog IF_MVAR(,ls) (Dialog title controls atts) IF_MVAR(,ps)
		= do {
			IF_MVAR(,ps1 <-) dOpen windowFunctions IF_MVAR(,ps);
			maybe_okId <- validateWindowId maybe_id;
			case maybe_okId of
				Nothing   -> throwGUI ErrorIdsInUse
				Just okId -> do
					cs   <- controlToHandles controls
					it   <- ioStGetIdTable
					rt   <- ioStGetReceiverTable
					ioId <- accIOEnv ioStGetIOId
					let itemHs          = map controlStateToWElementHandle cs
				    	(case controlIdsAreConsistent ioId okId itemHs rt it of
				    		Nothing -> throwGUI ErrorIdsInUse
				    		Just (it,rt1) ->
				    			let it1 = snd (addIdToIdTable okId (IdParent {idpIOId=ioId,idpDevice=WindowDevice,idpId=okId}) it)
				    			in do				    	
								ioStSetIdTable it1
								ioStSetReceiverTable rt1
								openModalWindow okId (WindowLSHandle {wlsState=ls,wlsHandle=initWindowHandle title Modal IsDialog NoWindowInfo itemHs atts}) ps1)
		}
		where
			maybe_id = getWindowIdAttribute atts


getWindowIdAttribute :: [WindowAttribute IF_MVAR(,ls ps)] -> Maybe Id
getWindowIdAttribute atts
	| hasIdAtt  = Just (getWindowIdAtt idAtt)
	| otherwise = Nothing
	where
		(hasIdAtt,idAtt) = cselect isWindowId undefined atts


{-	closeWindow closes the indicated window.
-}
closeWindow :: Id -> IF_MVAR(GUI (),ps -> GUI ps ps)
closeWindow id IF_MVAR(,ps)
	= disposeWindow (toWID id) IF_MVAR(,ps)

closeActiveWindow :: IF_MVAR(GUI (),ps -> GUI ps ps)
closeActiveWindow IF_MVAR(,ps)
	= do {
		maybeId <- getActiveWindow;
		if   isNothing maybeId
		then return IF_MVAR((),ps)
		else closeWindow (fromJust maybeId) IF_MVAR(,ps)
	  }

{-
setActiveWindow :: !Id !(PSt .l) -> PSt .l
setActiveWindow wId pState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice pState.io
	| not found
		= {pState & io=ioState}
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (exists,windows)			= hasWindowHandlesWindow wid windows
	| not exists				// Indicated window does not exist
		= {pState & io=IOStSetDevice (WindowSystemState windows) ioState}
	# (activeWIDS,windows)		= getWindowHandlesActiveWindow windows
	| isNothing activeWIDS		// There are no windows, so skip it
		= {pState & io=IOStSetDevice (WindowSystemState windows) ioState}
	# wids						= fromJust activeWIDS
	| exists && wids.wId==wId	// If already active, then skip
		= {pState & io=IOStSetDevice (WindowSystemState windows) ioState}
	# wHs						= windows.whsWindows
	  (modal,modeless)			= Uspan ismodalwindow wHs
	  (isModal,modal)			= UContains (identifyWindowStateHandle wid) modal
	| isModal					// Modal windows should not be activated
		= {pState & io=IOStSetDevice (WindowSystemState {windows & whsWindows=modal++modeless}) ioState}
	# (osdInfo,ioState)			= IOStGetOSDInfo ioState
	  isSDI						= getOSDInfoDocumentInterface osdInfo==SDI
	  (framePtr,clientPtr)		= case (getOSDInfoOSInfo osdInfo) of
	  								Just info -> (info.osFrame,info.osClient)
	  								_         -> (OSNoWindowPtr,OSNoWindowPtr)
	| isEmpty modal				// There are no modal windows, so put activated window in front
		# (_,wsH,others)		= URemove (identifyWindowStateHandle wid) undef modeless
		  (shown,wsH)			= getWindowStateHandleShow wsH
		# (wids,wsH)			= getWindowStateHandleWIDS wsH
		  activatePtr			= if (isSDI && wids.wPtr==clientPtr) framePtr wids.wPtr		// Do not activate SDI client, but SDI frame
		  showAction			= if shown id (snd o OSshowWindow activatePtr True)
		# ioState				= IOStSetDevice (WindowSystemState {windows & whsWindows=[wsH:others]}) ioState
//		# (delayinfo,ioState)	= accIOToolbox (OSactivateWindow osdInfo activatePtr o showAction) ioState
		# (tb,ioState)			= getIOToolbox ioState
		# pState				= {pState & io=ioState}
		# (delayinfo,pState,tb)	= OSactivateWindow osdInfo activatePtr handleOSEvent pState (showAction tb)
		# ioState				= setIOToolbox tb pState.io
		# ioState				= bufferDelayedEvents delayinfo ioState
		= {pState & io=ioState}
	| otherwise					// There are modal windows, so put activated window behind last modal
		# (befModals,lastModal)	= InitLast modal
		  (modalWIDS,lastModal)	= getWindowStateHandleWIDS lastModal
		  (_,wsH,others)		= URemove (identifyWindowStateHandle wid) undef modeless
		  (shown,wsH)			= getWindowStateHandleShow wsH
		  (modelessWIDS,wsH)	= getWindowStateHandleWIDS wsH
		  activatePtr			= if (isSDI && modelessWIDS.wPtr==clientPtr) framePtr modelessWIDS.wPtr	// Do not activate SDI client, but SDI frame
		  showAction			= if shown id (snd o OSshowWindow activatePtr True)
		# ioState				= IOStSetDevice (WindowSystemState {windows & whsWindows=befModals++[lastModal,wsH:others]}) ioState
//		# ioState				= appIOToolbox (showAction o OSstackWindow activatePtr modalWIDS.wPtr) ioState
		# (tb,ioState)			= getIOToolbox ioState
		# pState				= {pState & io=ioState}
		# (delayinfo,pState,tb)	= OSstackWindow activatePtr modalWIDS.wPtr handleOSEvent pState (showAction tb)
		# ioState				= setIOToolbox tb pState.io
		# ioState				= bufferDelayedEvents delayinfo ioState
		= {pState & io=ioState}
where
	wid							= toWID wId
	
	ismodalwindow :: !(WindowStateHandle .pst) -> (!Bool,!WindowStateHandle .pst)
	ismodalwindow wsH
		# (mode,wsH)			= getWindowStateHandleWindowMode wsH
		= (mode==Modal,wsH)
-}

{-	getActiveWindow returns the Id of the currently active window.
-}
getActiveWindow :: GUI IF_MVAR(,ps) (Maybe Id)
getActiveWindow
	= do {
		(found,wDevice) <- accIOEnv (ioStGetDevice WindowDevice);
		if   not found
		then return Nothing
		else let
			windows    = windowSystemStateGetWindowHandles wDevice
			activeWIDS = getWindowHandlesActiveWindow windows
		in do {
			appIOEnv (ioStSetDevice (WindowSystemState windows));
			return (fmap wId activeWIDS)
		   }
	  }
	  
{-

/*	setActiveControl makes the indicated control active only if its parent window is already active.
*/
setActiveControl :: !Id !(PSt .l) -> PSt .l
setActiveControl controlId pState=:{io}
	# (parentId,ioState)		= getParentId controlId io
	| isNothing parentId
		= {pState & io=ioState}
	# (activeId,ioState)		= getActiveWindow ioState
	| isNothing activeId
		= {pState & io=ioState}
	# parentId					= fromJust parentId
	# activeId					= fromJust activeId
	| parentId<>activeId
		= {pState & io=ioState}
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= {pState & io=ioState}
	# windows					= WindowSystemStateGetWindowHandles wDevice
	# (found,wsH,windows)		= getWindowHandlesWindow (toWID activeId) windows
	| not found
		= StdWindowFatalError "setActiveControl" "parent window could not be located"
	| otherwise
		# (tb,ioState)			= getIOToolbox ioState
		# (delayinfo,wsH,tb)	= setactivecontrol controlId wsH tb
		# ioState				= setIOToolbox tb ioState
		  windows				= setWindowHandlesWindow wsH windows
		# ioState				= IOStSetDevice (WindowSystemState windows) ioState
		# ioState				= bufferDelayedEvents delayinfo ioState
		= {pState & io=ioState}
where
	setactivecontrol :: !Id !(WindowStateHandle .pst) !*OSToolbox -> (![DelayActivationInfo],!WindowStateHandle .pst,!*OSToolbox)
	setactivecontrol controlId wsH=:{wshIds,wshHandle=Just wlsH=:{wlsHandle=wH=:{whKeyFocus,whItems}}} tb
		# (found,itemNr,itemPtr,itemHs)	= getWElementHandlesItemNrPtr controlId whItems
		| not found
			= StdWindowFatalError "setActiveControl" "indicated control could not be located"
		| otherwise
		//	# keyfocus					= setNewFocusItem itemNr whKeyFocus
			# (delayinfo,tb)			= OSactivateControl wshIds.wPtr itemPtr tb
			= (delayinfo,{wsH & wshHandle=Just {wlsH & wlsHandle={wH & /*whKeyFocus=keyfocus,*/whItems=itemHs}}},tb)
	where
		getWElementHandlesItemNrPtr :: !Id ![WElementHandle .ls .pst] -> (!Bool,!Int,!OSWindowPtr,![WElementHandle .ls .pst])
		getWElementHandlesItemNrPtr id [itemH:itemHs]
			# (found,itemNr,itemPtr,itemH)	= getWElementHandleItemNrPtr id itemH
			| found
				= (found,itemNr,itemPtr,[itemH:itemHs])
			| otherwise
				# (found,itemNr,itemPtr,itemHs)	= getWElementHandlesItemNrPtr id itemHs
				= (found,itemNr,itemPtr,[itemH:itemHs])
		where
			getWElementHandleItemNrPtr :: !Id !(WElementHandle .ls .pst) -> (!Bool,!Int,!OSWindowPtr,!WElementHandle .ls .pst)
			getWElementHandleItemNrPtr id (WItemHandle itemH=:{wItemNr,wItems,wItemId,wItemPtr})
				| isNothing wItemId || fromJust wItemId<>id
					# (found,itemNr,itemPtr,itemHs)	= getWElementHandlesItemNrPtr id wItems
					= (found,itemNr,itemPtr,WItemHandle {itemH & wItems=itemHs})
				| otherwise
					= (True,wItemNr,wItemPtr,WItemHandle itemH)
			getWElementHandleItemNrPtr itemNr (WListLSHandle itemHs)
				# (found,itemNr,itemPtr,itemHs)		= getWElementHandlesItemNrPtr id itemHs
				= (found,itemNr,itemPtr,WListLSHandle itemHs)
			getWElementHandleItemNrPtr itemNr (WExtendLSHandle wExH=:{wExtendItems=itemHs})
				# (found,itemNr,itemPtr,itemHs)		= getWElementHandlesItemNrPtr id itemHs
				= (found,itemNr,itemPtr,WExtendLSHandle {wExH & wExtendItems=itemHs})
			getWElementHandleItemNrPtr itemNr (WChangeLSHandle wChH=:{wChangeItems=itemHs})
				# (found,itemNr,itemPtr,itemHs)		= getWElementHandlesItemNrPtr id itemHs
				= (found,itemNr,itemPtr,WChangeLSHandle {wChH & wChangeItems=itemHs})
		getWElementHandlesItemNrPtr _ _
			= (False,0,OSNoWindowPtr,[])
	setactivecontrol _ _ _
		= StdWindowFatalError "setActiveControl" "unexpected window placeholder argument"

/*	getActiveControl returns the Id of the currently active control. The Bool result is True only iff the 
	control could be found. In that case, if the control had an Id attribute, then (Just id) is returned,
	Nothing otherwise.
*/
getActiveControl :: !(IOSt .l) -> (!(!Bool,!Maybe Id),!IOSt .l)
getActiveControl ioState
	# (activeId,ioState)		= getActiveWindow ioState
	| isNothing activeId
		= ((False,Nothing),ioState)
	# activeId					= fromJust activeId
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= ((False,Nothing),ioState)
	# windows					= WindowSystemStateGetWindowHandles wDevice
	# (hasWindow,wsH,windows)	= getWindowHandlesWindow (toWID activeId) windows
	| not hasWindow
		= StdWindowFatalError "getActiveControl" "active window could not be located"
	# (keyfocus,wsH)			= getWindowStateHandleKeyFocus wsH
	  maybeItemNr				= getCurrentFocusItem keyfocus
	| isNothing maybeItemNr
		# windows				= setWindowHandlesWindow wsH windows
		# ioState				= IOStSetDevice (WindowSystemState windows) ioState
		= ((False,Nothing),ioState)
	# (foundId,wsH)				= getControlIdFromItemNr (fromJust maybeItemNr) wsH
	  windows					= setWindowHandlesWindow wsH windows
	# ioState					= IOStSetDevice (WindowSystemState windows) ioState
	= (foundId,ioState)
where
	getControlIdFromItemNr :: !Int !(WindowStateHandle .pst) -> (!(!Bool,!Maybe Id),!WindowStateHandle .pst)
	getControlIdFromItemNr itemNr wsH=:{wshHandle=Just wlsH=:{wlsHandle=wH=:{whItems}}}
		# (_,foundId,itemHs)	= getWElementHandlesIdFromItemNr itemNr whItems
		= (foundId,{wsH & wshHandle=Just {wlsH & wlsHandle={wH & whItems=itemHs}}})
	where
		getWElementHandlesIdFromItemNr :: !Int ![WElementHandle .ls .pst] -> (!Bool,!(!Bool,!Maybe Id),![WElementHandle .ls .pst])
		getWElementHandlesIdFromItemNr itemNr [itemH:itemHs]
			# (found,foundId,itemH)		= getWElementHandleIdFromItemNr itemNr itemH
			| found
				= (found,foundId,[itemH:itemHs])
			| otherwise
				# (found,foundId,itemHs)= getWElementHandlesIdFromItemNr itemNr itemHs
				= (found,foundId,[itemH:itemHs])
		where
			getWElementHandleIdFromItemNr :: !Int !(WElementHandle .ls .pst) -> (!Bool,!(!Bool,!Maybe Id),!WElementHandle .ls .pst)
			getWElementHandleIdFromItemNr itemNr (WItemHandle itemH=:{wItemNr,wItems,wItemId})
				| itemNr<>wItemNr
					# (found,foundId,itemHs)	= getWElementHandlesIdFromItemNr itemNr wItems
					= (found,foundId,WItemHandle {itemH & wItems=itemHs})
				| otherwise
					= (True,(True,wItemId),WItemHandle itemH)
			getWElementHandleIdFromItemNr itemNr (WListLSHandle itemHs)
				# (found,foundId,itemHs)		= getWElementHandlesIdFromItemNr itemNr itemHs
				= (found,foundId,WListLSHandle itemHs)
			getWElementHandleIdFromItemNr itemNr (WExtendLSHandle wExH=:{wExtendItems=itemHs})
				# (found,foundId,itemHs)		= getWElementHandlesIdFromItemNr itemNr itemHs
				= (found,foundId,WExtendLSHandle {wExH & wExtendItems=itemHs})
			getWElementHandleIdFromItemNr itemNr (WChangeLSHandle wChH=:{wChangeItems=itemHs})
				# (found,foundId,itemHs)		= getWElementHandlesIdFromItemNr itemNr itemHs
				= (found,foundId,WChangeLSHandle {wChH & wChangeItems=itemHs})
		getWElementHandlesIdFromItemNr _ _
			= (False,(False,Nothing),[])
	getControlIdFromItemNr _ _
		= StdWindowFatalError "getActiveControl" "unexpected window placeholder argument"


/*	stackWindow changes the stacking order of the current windows. */

stackWindow :: !Id !Id !(PSt .l) -> PSt .l
stackWindow windowId behindId pState=:{io=ioState}
	| windowId==behindId	// Don't stack a window behind itself
		= pState
	# (found,wDevice,ioState)		= IOStGetDevice WindowDevice ioState
	| not found
		= {pState & io=ioState}
	# windows						= WindowSystemStateGetWindowHandles wDevice
	# (hasBehind,windows)			= hasWindowHandlesWindow behindWID windows
	| not hasBehind			// Behind window does not exist
		# ioState					= IOStSetDevice (WindowSystemState windows) ioState
		= {pState & io=ioState}
	# (hasWindow,wsH,windows)		= getWindowHandlesWindow windowWID windows
	| not hasWindow			// Stack window does not exist
		# ioState					= IOStSetDevice (WindowSystemState windows) ioState
		= {pState & io=ioState}
	# (mode,wsH)					= getWindowStateHandleWindowMode wsH
	| mode==Modal			// Stack window is modal, skip
		# ioState					= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
		= {pState & io=ioState}
	| otherwise
		# (_,_,windows)				= removeWindowHandlesWindow windowWID windows		// remove placeholder window
		# (wids,wsH)				= getWindowStateHandleWIDS wsH
		# (behindWIDS,windows)		= addBehindWindowHandlesWindow behindWID wsH windows
		# ioState					= IOStSetDevice (WindowSystemState windows) ioState
		# (tb,ioState)				= getIOToolbox ioState
		# pState					= {pState & io=ioState}
		# (delayinfo,pState,tb)		= OSstackWindow wids.wPtr behindWIDS.wPtr handleOSEvent pState tb
		# ioState					= setIOToolbox tb pState.io
		# ioState					= bufferDelayedEvents delayinfo ioState
		= {pState & io=ioState}
where
	windowWID						= toWID windowId
	behindWID						= toWID behindId

/*	handleOSEvent turns handleOneEventForDevices into the form required by OSactivateWindow and OSstackWindow.
	(Used by stackWindow, setActiveWindow.)
*/
handleOSEvent :: !OSEvent !(!PSt .l,!*OSToolbox) -> (!PSt .l,!*OSToolbox)
handleOSEvent osEvent (pState,tb)
	= (thd3 (handleOneEventForDevices (ScheduleOSEvent osEvent []) pState),tb)

getWindowStack :: !(IOSt .l) -> (![(Id,WindowType)],!IOSt .l)
getWindowStack ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= ([],ioState)
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  wsHs						= windows.whsWindows
	  (id_types,wsHs)			= unzip (map getWindowIdType wsHs)
	= (id_types,IOStSetDevice (WindowSystemState {windows & whsWindows=wsHs}) ioState)
where
	getWindowIdType :: !(WindowStateHandle .pst) -> ((Id,WindowType),!WindowStateHandle .pst)
	getWindowIdType wsH
		# (wids,wsH)	= getWindowStateHandleWIDS wsH
		# (kind,wsH)	= getWindowStateHandleWindowKind wsH
		= ((wids.wId,if (kind==IsWindow) windowtype dialogtype),wsH)

getWindowsStack :: !(IOSt .l) -> (![Id],!IOSt .l)
getWindowsStack ioState
	# (id_types,ioState)	= getWindowStack ioState
	= (FilterMap (\(id,wtype)->(wtype==windowtype,id)) id_types,ioState)

getDialogsStack :: !(IOSt .l) -> (![Id],!IOSt .l)
getDialogsStack ioState
	# (id_types,ioState)	= getWindowStack ioState
	= (FilterMap (\(id,wtype)->(wtype==dialogtype,id)) id_types,ioState)


/*	Return layout attributes and default values.
*/
getDefaultHMargin :: !Bool !(IOSt .l) -> ((Int,Int),!IOSt .l)
getDefaultHMargin isWindow ioState
	| isWindow
		= ((0,0),ioState)
	| otherwise
		# ({osmHorMargin},ioState)	= IOStGetOSWindowMetrics ioState
		= ((osmHorMargin,osmHorMargin),ioState)

getDefaultVMargin :: !Bool !(IOSt .l) -> ((Int,Int),!IOSt .l)
getDefaultVMargin isWindow ioState
	| isWindow
		= ((0,0),ioState)
	| otherwise
		# ({osmVerMargin},ioState)	= IOStGetOSWindowMetrics ioState
		= ((osmVerMargin,osmVerMargin),ioState)

getDefaultItemSpace :: !Bool !(IOSt .l) -> ((Int,Int),!IOSt .l)
getDefaultItemSpace _ ioState
	# ({osmHorItemSpace,osmVerItemSpace},ioState)	= IOStGetOSWindowMetrics ioState
	= ((osmHorItemSpace,osmVerItemSpace),ioState)

getWindowHMargin :: !Id	!(IOSt .l) -> (!Maybe (Int,Int),!IOSt .l)
getWindowHMargin id ioState
	# (found,wDevice,ioState)		= IOStGetDevice WindowDevice ioState
	| not found
		= (Nothing,ioState)
	# windows						= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)			= getWindowHandlesWindow (toWID id) windows
	| not found
		= (Nothing,IOStSetDevice (WindowSystemState windows) ioState)
	| otherwise
		# (wMetrics,ioState)		= IOStGetOSWindowMetrics ioState
		# (marginAtt,wsH)			= gethmargin wMetrics wsH
		= (Just marginAtt,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)
where
	gethmargin :: !OSWindowMetrics !(WindowStateHandle .pst) -> ((Int,Int),!WindowStateHandle .pst)
	gethmargin wMetrics wsH=:{wshHandle=Just {wlsHandle={whKind,whAtts}}}
		= (getWindowHMargins whKind wMetrics whAtts,wsH)
	gethmargin _ _
		= StdWindowFatalError "getWindowHMargin" "unexpected window placeholder argument"

getWindowVMargin :: !Id	!(IOSt .l) -> (!Maybe (Int,Int),!IOSt .l)
getWindowVMargin id ioState
	# (found,wDevice,ioState)		= IOStGetDevice WindowDevice ioState
	| not found
		= (Nothing,ioState)
	# windows						= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)			= getWindowHandlesWindow (toWID id) windows
	| not found
		= (Nothing,IOStSetDevice (WindowSystemState windows) ioState)
	| otherwise
		# (wMetrics,ioState)		= IOStGetOSWindowMetrics ioState
		# (marginAtt,wsH)			= getvmargin wMetrics wsH
		= (Just marginAtt,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)
where
	getvmargin :: !OSWindowMetrics !(WindowStateHandle .pst) -> ((Int,Int),!WindowStateHandle .pst)
	getvmargin wMetrics wsH=:{wshHandle=Just {wlsHandle={whKind,whAtts}}}
		= (getWindowVMargins whKind wMetrics whAtts,wsH)
	getvmargin _ _
		= StdWindowFatalError "getWindowVMargin" "unexpected window placeholder argument"

getWindowItemSpace :: !Id !(IOSt .l) -> (!Maybe (Int,Int),!IOSt .l)
getWindowItemSpace id ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= (Nothing,ioState)
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= (Nothing,IOStSetDevice (WindowSystemState windows) ioState)
	| otherwise
		# ({osmHorItemSpace,osmVerItemSpace},ioState)
								= IOStGetOSWindowMetrics ioState
		# (marginAtt,wsH)		= getitemspaces (osmHorItemSpace,osmVerItemSpace) wsH
		= (Just marginAtt,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)
where
	getitemspaces :: (Int,Int) !(WindowStateHandle .pst) -> ((Int,Int),!WindowStateHandle .pst)
	getitemspaces (defHSpace,defVSpace) wsH=:{wshHandle=Just {wlsHandle={whAtts}}}
		= (getWindowItemSpaceAtt (snd (Select isWindowItemSpace (WindowItemSpace defHSpace defVSpace) whAtts)),wsH)
	getitemspaces _ _
		= StdWindowFatalError "getWindowItemSpace" "unexpected window placeholder argument"
-}


--	Setting the SelectState of windows.
{-
enableWindow :: Id -> GUI ps ()
enableWindow id = do
	(found,wDevice)	= accIOEnv (ioStGetDevice WindowDevice)
	if not found then return ()
	else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	     in if not found || getWindowStateHandleWindowKind wsH/=IsWindow || getWindowStateHandleSelect wsH
	        then return ()
	        else do	
			osdInfo <- accIOEnv ioStGetOSDInfo
		  	let framePtr = case getOSDInfoOSInfo osdInfo of
		  		Just info -> osFrame info
		  		_         -> osNoWindowPtr
			wMetrics <- accIOEnv ioStGetOSWindowMetrics
		  	let wsH1 = setWindowStateHandleSelect True wsH
		  	let wids = getWindowStateHandleWIDS wsH
		# (wH`,wsH,tb)			= retrieveWindowHandle` wsH tb
		# (wH`,tb)				= enablecontrols [] True wMetrics (wPtr wids) wH` tb
		  wsH					= insertWindowHandle` wH` wsH
		  (windowInfo,wsH)		= getWindowStateHandleWindowInfo wsH
		  scrollInfo			= case windowInfo of
		  							WindowInfo info	-> (isJust info.windowHScroll,isJust info.windowVScroll)
		  							other			-> (False,False)
		# tb					= OSenableWindow (if getOSDInfoDocumentInterface osdInfo==SDI then framePtr else (wPtr wids)) scrollInfo False tb
		# tb					= OSinvalidateWindow (wPtr wids) tb
		# ioState				= setIOToolbox tb ioState
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState

disableWindow :: !Id !(IOSt .l) -> IOSt .l
disableWindow id ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= ioState
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= IOStSetDevice (WindowSystemState windows) ioState
	# (wKind,wsH)				= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	# (curSelectState,wsH)		= getWindowStateHandleSelect wsH
	| not curSelectState
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	| otherwise
		# (osdInfo, ioState)	= IOStGetOSDInfo ioState
		  isSDI					= getOSDInfoDocumentInterface osdInfo==SDI
		  framePtr				= case (getOSDInfoOSInfo osdInfo) of
		  							Just info -> info.osFrame
		  							_         -> OSNoWindowPtr
		# (wMetrics,ioState)	= IOStGetOSWindowMetrics ioState
		# (tb,ioState)			= getIOToolbox ioState
		  wsH					= setWindowStateHandleSelect False wsH
		  (wids,wsH)			= getWindowStateHandleWIDS wsH
		  wPtr					= wids.wPtr
		# (wH`,wsH,tb)			= retrieveWindowHandle` wsH tb
		# (wH`,tb)				= disablecontrols [] True wMetrics wPtr wH` tb
		  wsH					= insertWindowHandle` wH` wsH
		/* Mike
		  (maybeWindowInfo,wsH)	= getWindowStateHandleWindowInfo wsH
		  scrollInfo			= case maybeWindowInfo of
		  							Nothing		-> (False,False)
		  							Just info	-> (isJust info.windowHScroll,isJust info.windowVScroll)
		*/
		  (windowInfo,wsH)		= getWindowStateHandleWindowInfo wsH
		  scrollInfo			= case windowInfo of
		  							WindowInfo info	-> (isJust info.windowHScroll,isJust info.windowVScroll)
		  							other			-> (False,False)
		# tb					= OSdisableWindow (if isSDI framePtr wPtr) scrollInfo False tb
		# tb					= OSinvalidateWindow wPtr tb
		# ioState				= setIOToolbox tb ioState
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
-}

enableWindowMouse  :: Id -> GUI ps ()
enableWindowMouse  id = setWindowMouseSelectState Able   id

disableWindowMouse :: Id -> GUI ps ()
disableWindowMouse id = setWindowMouseSelectState Unable id

setWindowMouseSelectState :: SelectState -> Id -> GUI ps ()
setWindowMouseSelectState selectState id = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return ()
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in
		if not found
		then appIOEnv (ioStSetDevice (WindowSystemState windows))
		else if getWindowStateHandleWindowKind wsH /= IsWindow
		     then appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
		     else
			let wsH1 = setMouseSelectState selectState wsH
			in appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH1 windows1))))
	where
		setMouseSelectState :: SelectState -> WindowStateHandle ps -> WindowStateHandle ps
		setMouseSelectState selectState (WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH}))) =
			WindowStateHandle wids (Just wlsH{wlsHandle=wH{whAtts=setMouseSelectStateAtt selectState (whAtts wH)}})
			where
				setMouseSelectStateAtt :: SelectState -> [WindowAttribute ls ps] -> [WindowAttribute ls ps]
				setMouseSelectStateAtt selectState atts
					| not found = atts
					| otherwise =
						let (filter,_,fun) = getWindowMouseAtt mouseAtt
						in (WindowMouse filter selectState fun:atts1)
					where
						(found,mouseAtt,atts1) = remove isWindowMouse undefined atts
		setMouseSelectState _ _ = stdWindowFatalError "setWindowMouseSelectState" "unexpected window placeholder argument"


enableWindowKeyboard  :: Id -> GUI ps ()
enableWindowKeyboard  id = setWindowKeyboardSelectState Able   id

disableWindowKeyboard :: Id -> GUI ps ()
disableWindowKeyboard id = setWindowKeyboardSelectState Unable id

setWindowKeyboardSelectState :: SelectState -> Id -> GUI ps ()
setWindowKeyboardSelectState selectState id = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return ()
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in if not found || getWindowStateHandleWindowKind wsH /= IsWindow then return ()
	         else do
			let wsH1 = setKeyboardSelectState selectState wsH
			appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH1 windows1))))
	where
		setKeyboardSelectState :: SelectState -> WindowStateHandle ps -> WindowStateHandle ps
		setKeyboardSelectState selectState (WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH}))) =
			WindowStateHandle wids (Just wlsH{wlsHandle=wH{whAtts=setKeyboardSelectStateAtt selectState (whAtts wH)}})
			where
				setKeyboardSelectStateAtt :: SelectState -> [WindowAttribute ls ps] -> [WindowAttribute ls ps]
				setKeyboardSelectStateAtt selectState atts					
					| not found = atts
					| otherwise =
						let (filter,_,fun) = getWindowKeyboardAtt keyAtt
						in (WindowKeyboard filter selectState fun:atts1)
					where
						(found,keyAtt,atts1) = remove isWindowKeyboard undefined atts
		setKeyboardSelectState _ _ =
			stdWindowFatalError "setWindowKeyboardSelectState" "unexpected window placeholder argument"



{-
getWindowSelectState :: !Id !(IOSt .l) -> (!Maybe SelectState,!IOSt .l)
getWindowSelectState id ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= (Nothing,ioState)
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= (Nothing,IOStSetDevice (WindowSystemState windows) ioState)
	# (wKind,wsH)				= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= (Nothing,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)
	| otherwise
		# (wSelect,wsH)			= getWindowStateHandleSelect wsH
		= (Just (if wSelect Able Unable),IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)

getWindowMouseSelectState :: !Id !(IOSt .l) -> (!Maybe SelectState,!IOSt .l)
getWindowMouseSelectState id ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= (Nothing,ioState)
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= (Nothing,IOStSetDevice (WindowSystemState windows) ioState)
	# (wKind,wsH)				= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= (Nothing,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)
	| otherwise
		# (found,_,select,wsH)	= getWindowMouseAttInfo wsH
		= (if found (Just select) Nothing,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)

getWindowMouseAttInfo :: !(WindowStateHandle .pst) -> (!Bool,MouseStateFilter,SelectState,!WindowStateHandle .pst)
getWindowMouseAttInfo wsH=:{wshHandle=Just wlsH=:{wlsHandle=wH=:{whAtts}}}
	# (hasMouseAtt,mouseAtt)	= Select isWindowMouse undef whAtts
	| not hasMouseAtt
		= (False,undef,undef,wsH)
	| otherwise
		# (filter,selectState,_)= getWindowMouseAtt mouseAtt
		= (True,filter,selectState,wsH)
getWindowMouseAttInfo _
	= StdWindowFatalError "getWindowMouseAttInfo" "unexpected window placeholder argument"
-}
getWindowKeyboardSelectState :: Id -> GUI ps (Maybe SelectState)
getWindowKeyboardSelectState id = do
	(found,wDevice) <- accIOEnv (ioStGetDevice WindowDevice);
	(if not found then return Nothing
	 else
	   let windows = windowSystemStateGetWindowHandles wDevice
	       (found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	   in
	     if not found then appIOEnv (ioStSetDevice (WindowSystemState windows)) >> return Nothing
	     else do
	        appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
		(if getWindowStateHandleWindowKind wsH /= IsWindow 
		 then return Nothing
		 else case getWindowKeyboardAttInfo wsH of
	  		Just (_,select) -> return (Just select)
	  		Nothing 	-> return Nothing))

getWindowKeyboardAttInfo :: WindowStateHandle ps -> Maybe (KeyboardStateFilter,SelectState)
getWindowKeyboardAttInfo wsH@(WindowStateHandle wids' (Just (WindowLSHandle {wlsHandle=wH@(WindowHandle {whAtts=whAtts})}))) =
	case find isWindowKeyboard whAtts of
	  Just keyAtt -> let (filter,selectState,_)= getWindowKeyboardAtt keyAtt
	  		 in Just (filter,selectState)
	  Nothing -> Nothing
getWindowKeyboardAttInfo _
	= stdWindowFatalError "getWindowKeyboardAttInfo" "unexpected window placeholder argument"

{-
getWindowMouseStateFilter :: !Id !(IOSt .l) -> (!Maybe MouseStateFilter,!IOSt .l)
getWindowMouseStateFilter id ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= (Nothing,ioState)
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= (Nothing,IOStSetDevice (WindowSystemState windows) ioState)
	# (wKind,wsH)				= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= (Nothing,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)
	| otherwise
		# (found,filter,_,wsH)	= getWindowMouseAttInfo wsH
		= (if found (Just filter) Nothing,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)

getWindowKeyboardStateFilter :: !Id !(IOSt .l) -> (!Maybe KeyboardStateFilter,!IOSt .l)
getWindowKeyboardStateFilter id ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= (Nothing,ioState)
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= (Nothing,IOStSetDevice (WindowSystemState windows) ioState)
	# (wKind,wsH)				= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= (Nothing,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)
	| otherwise
		# (found,filter,_,wsH)	= getWindowKeyboardAttInfo wsH
		= (if found (Just filter) Nothing,IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState)

setWindowMouseStateFilter :: !Id !MouseStateFilter !(IOSt .l) -> IOSt .l
setWindowMouseStateFilter id filter ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= ioState
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= IOStSetDevice (WindowSystemState windows) ioState
	# (wKind,wsH)				= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	| otherwise
		# wsH					= setMouseFilter filter wsH
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
where
	setMouseFilter :: !MouseStateFilter !(WindowStateHandle .pst) -> WindowStateHandle .pst
	setMouseFilter filter wsH=:{wshHandle=Just wlsH=:{wlsHandle=wH=:{whAtts}}}
		= {wsH & wshHandle=Just {wlsH & wlsHandle={wH & whAtts=setMouseStateFilterAtt filter whAtts}}}
	where
		setMouseStateFilterAtt :: !MouseStateFilter ![WindowAttribute .pst] -> [WindowAttribute .pst]
		setMouseStateFilterAtt filter atts
			# (found,mouseAtt,atts)	= Remove isWindowMouse undef atts
			| not found
				= atts
			| otherwise
				# (_,select,fun)	= getWindowMouseAtt mouseAtt
				= [WindowMouse filter select fun:atts]
	setMouseFilter _ _
		= StdWindowFatalError "setWindowMouseStateFilter" "unexpected window placeholder argument"

setWindowKeyboardStateFilter :: !Id !KeyboardStateFilter !(IOSt .l) -> IOSt .l
setWindowKeyboardStateFilter id filter ioState
	# (found,wDevice,ioState)		= IOStGetDevice WindowDevice ioState
	| not found
		= ioState
	# windows						= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)			= getWindowHandlesWindow (toWID id) windows
	| not found
		= IOStSetDevice (WindowSystemState windows) ioState
	# (wKind,wsH)					= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	| otherwise
		# wsH						= setKeyboardFilter filter wsH
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
where
	setKeyboardFilter :: !KeyboardStateFilter !(WindowStateHandle .pst) -> WindowStateHandle .pst
	setKeyboardFilter filter wsH=:{wshHandle=Just wlsH=:{wlsHandle=wH=:{whAtts}}}
		= {wsH & wshHandle=Just {wlsH & wlsHandle={wH & whAtts=setKeyboardStateFilterAtt filter whAtts}}}
	where
		setKeyboardStateFilterAtt :: !KeyboardStateFilter ![WindowAttribute .pst] -> [WindowAttribute .pst]
		setKeyboardStateFilterAtt filter atts
			# (found,keyAtt,atts)	= Remove isWindowKeyboard undef atts
			| not found
				= atts
			| otherwise
				# (_,select,fun)	= getWindowKeyboardAtt keyAtt
				= [WindowKeyboard filter select fun:atts]
	setKeyboardFilter _ _
		= StdWindowFatalError "setWindowKeyboardStateFilter" "unexpected window placeholder argument"
-}

drawInWindow :: Id -> Draw x -> GUI ps (Maybe x)
drawInWindow id drawf = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return Nothing
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in if not found then do
	     		appIOEnv (ioStSetDevice (WindowSystemState windows))
	     		return Nothing
	      	 else if getWindowStateHandleWindowKind wsH /= IsWindow
	     	      then do
	     	     	appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
	     	        return Nothing
	              else do
	             	wMetrics <- accIOEnv ioStGetOSWindowMetrics
			(x,wsH)	<- liftIO (drawInWindow' wMetrics drawf wsH)
			appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
			return (Just x))
	where
		drawInWindow' :: OSWindowMetrics -> Draw x -> WindowStateHandle ps -> IO (x,WindowStateHandle ps)
		drawInWindow' wMetrics drawf (WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH}))) = do
			wH <- validateWindowClipState wMetrics False (wPtr wids) wH
			(x,wH) <- WD.drawInWindow wMetrics (wPtr wids) drawf wH
			return (x, WindowStateHandle wids (Just wlsH{wlsHandle=wH}))
		drawInWindow' _ _ _ = stdWindowFatalError "drawInWindow" "unexpected window placeholder argument"
{-
updateWindow :: !Id !(Maybe ViewFrame) !(IOSt .l) -> IOSt .l
updateWindow id maybeViewFrame ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= ioState
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= IOStSetDevice (WindowSystemState windows) ioState
	# (wKind,wsH)				= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	| otherwise
		# (wMetrics,ioState)	= IOStGetOSWindowMetrics ioState
		# (wsH,ioState)			= accIOToolbox (updateWindowBackground wMetrics maybeViewFrame wsH) ioState
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
where
	updateWindowBackground :: !OSWindowMetrics !(Maybe ViewFrame) !(WindowStateHandle .pst) !*OSToolbox
															   -> (!WindowStateHandle .pst, !*OSToolbox)
	updateWindowBackground wMetrics maybeViewFrame wsH=:{wshIds,wshHandle=Just wlsH=:{wlsHandle=wH}} tb
		| IsEmptyRect updArea
			= (wsH,tb)
		| otherwise
			# (wH,tb)					= updatewindow wMetrics updInfo wH tb
			= ({wsH & wshHandle=Just {wlsH & wlsHandle=wH}},tb)
	where
		whSize							= wH.whSize
		info							= getWindowInfoWindowData wH.whWindowInfo
		(origin,domainRect,hasScrolls)	= (info.windowOrigin,info.windowDomain,(isJust info.windowHScroll,isJust info.windowVScroll))
		visScrolls						= OSscrollbarsAreVisible wMetrics domainRect (toTuple whSize) hasScrolls
		contentRect						= getWindowContentRect wMetrics visScrolls (SizeToRect whSize)
		updArea							= case maybeViewFrame of
											Nothing		-> contentRect
											Just rect	-> IntersectRects (RectangleToRect (subVector (toVector origin) rect)) contentRect
		updInfo							= {	updWIDS			= wshIds
										  ,	updWindowArea	= updArea
										  ,	updControls		= []
										  ,	updGContext		= Nothing
										  }
	updateWindowBackground _ _ _ _
		= StdWindowFatalError "updateWindow" "unexpected window placeholder argument"
-}
setWindowLook :: Id -> Bool -> (Bool,Look) -> GUI ps ()
setWindowLook wId redraw (sysLook,lookFun) = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return ()
	 else do
		let windows	= windowSystemStateGetWindowHandles wDevice
	  	let (found,wsH,windows1) = getWindowHandlesWindow (toWID wId) windows
		(if not found
		 then appIOEnv (ioStSetDevice (WindowSystemState windows))
		 else if getWindowStateHandleWindowKind wsH /= IsWindow
		     then appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
		     else do
				wMetrics <- accIOEnv ioStGetOSWindowMetrics
				wsH <- liftIO (setwindowlook wMetrics redraw (sysLook,lookFun) wsH)
				appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))))
	where
		setwindowlook :: OSWindowMetrics -> Bool -> (Bool,Look) -> (WindowStateHandle ps) -> IO (WindowStateHandle ps)
		setwindowlook wMetrics redraw (sysLook,lookFun) wsH@(WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH}))) =
			let 
				lookInfo1 = lookInfo{lookFun=lookFun,lookSysUpdate=sysLook}		
				wH1 = wH{whWindowInfo=WindowInfo (windowInfo{windowLook=lookInfo1})}
			in if not redraw
			   then return (WindowStateHandle wids (Just wlsH{wlsHandle=wH1}))
			   else do
					wH2 <- validateWindowClipState wMetrics False (wPtr wids) wH1
					wH3 <- drawWindowLook wMetrics (wPtr wids) (return ()) updState wH2
					return (WindowStateHandle wids (Just wlsH{wlsHandle=wH3}))
			where
				windowInfo				= getWindowInfoWindowData (whWindowInfo wH)
				lookInfo				= windowLook windowInfo
				domainRect				= windowDomain windowInfo
				origin					= windowOrigin windowInfo
				hasScrolls				= (isJust (windowHScroll windowInfo),isJust (windowVScroll windowInfo))
				visScrolls				= osScrollbarsAreVisible wMetrics domainRect (toTuple (whSize wH)) hasScrolls
				contentRect				= getWindowContentRect wMetrics visScrolls (sizeToRect (whSize wH))
				wFrame					= posSizeToRectangle origin (rectSize contentRect)
				updState				= rectangleToUpdateState wFrame
		setwindowlook _ _ _ _ = stdWindowFatalError "setWindowLook" "unexpected window placeholder argument"

getWindowLook :: Id -> GUI ps (Maybe (Bool,Look))
getWindowLook id = do
	(found,wDevice) <- accIOEnv (ioStGetDevice WindowDevice)
	(if not found
	 then return Nothing
	 else let 
	 		windows = windowSystemStateGetWindowHandles wDevice
			(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in 
	      	if not found
		then do
			appIOEnv (ioStSetDevice (WindowSystemState windows))
			return Nothing
		else if getWindowStateHandleWindowKind wsH/=IsWindow
		     then do
		     	appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
			return Nothing
		     else
			let
				windowInfo = getWindowStateHandleWindowInfo wsH	
		  		LookInfo{lookFun=lookFun,lookSysUpdate=lookSysUpdate} = windowLook (getWindowInfoWindowData windowInfo)
			in do
				appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
				return (Just (lookSysUpdate,lookFun)))


--	Operations that are concerned with the position of windows/dialogues.
{-
setWindowPos :: !Id !ItemPos !(IOSt .l) -> IOSt .l
setWindowPos id pos ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= ioState
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= IOStSetDevice (WindowSystemState windows) ioState
	# (wMode,wsH)				= getWindowStateHandleWindowMode wsH
	| wMode==Modal
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	# (okId,pos,windows)		= validateRelativeId id pos windows
	| not okId
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	| otherwise
		# (wids, wsH)			= getWindowStateHandleWIDS wsH
		  (wSize,wsH)			= getWindowStateHandleSize wsH
		  (wKind,wsH)			= getWindowStateHandleWindowKind wsH
		  windows				= setWindowHandlesWindow wsH windows
		# (osdInfo, ioState)	= IOStGetOSDInfo ioState
		  isSDI					= getOSDInfoDocumentInterface osdInfo==SDI
		  (framePtr,clientPtr)	= case (getOSDInfoOSInfo osdInfo) of
		  							Just info -> (info.osFrame,info.osClient)
		  							_         -> (OSNoWindowPtr,OSNoWindowPtr)
		# (wMetrics,ioState)	= IOStGetOSWindowMetrics ioState
		# (tb,ioState)			= getIOToolbox ioState
		# (pos,windows,tb)		= exactWindowPos wMetrics wSize (Just pos) wKind Modeless windows tb
		# tb					= OSsetWindowPos (if (isSDI && wids.wPtr==clientPtr) framePtr wids.wPtr) (toTuple pos) True True tb
		# ioState				= setIOToolbox tb ioState
		= IOStSetDevice (WindowSystemState windows) ioState
where
	// validateRelativeId checks the validity of the ItemPos. 
	// It assumes that the WindowHandles argument is not empty (containing atleast the target window).
	validateRelativeId :: !Id !ItemPos !(WindowHandles .pst) -> (!Bool,!ItemPos,!WindowHandles .pst)
	validateRelativeId id itemPos=:(itemLoc,itemOffset) windows
		| isRelative
			# (exists,windows)	= hasWindowHandlesWindow (toWID relativeId) windows
			= (exists,itemPos,windows)
		| isRelativePrev
			# wsHs				= windows.whsWindows
			  (widsstack,wsHs)	= unzip (map getWindowStateHandleWIDS wsHs)
			  windows			= {windows & whsWindows=wsHs}
			  (found,prevId)	= findPrevId (toWID id) widsstack
			  itemLoc			= if (not found) itemLoc
			  					 (case itemLoc of
									LeftOfPrev  -> LeftOf  prevId
									RightToPrev -> RightTo prevId
									AbovePrev   -> Above   prevId
									BelowPrev   -> Below   prevId
								 )
			= (found,(itemLoc,itemOffset),windows)
			with
				findPrevId :: !WID ![WIDS] -> (!Bool,Id)
				findPrevId wid [_]
					= (False,undef)
				findPrevId wid [wid`:wids=:[wid``:_]]
					| identifyWIDS wid wid``	= (True,wid`.wId)
					| otherwise					= findPrevId wid [wid``:wids]
		| otherwise
			= (True,itemPos,windows)
	where
		(isRelative,relativeId)	= case itemLoc of
									LeftOf  id  -> (True,id)
									RightTo id  -> (True,id)
									Above   id  -> (True,id)
									Below   id  -> (True,id)
									_           -> (False,undef)
		isRelativePrev			= case itemLoc of
									LeftOfPrev  -> True
									RightToPrev -> True
									AbovePrev   -> True
									BelowPrev   -> True
									_           -> False

getWindowPos :: !Id !(IOSt .l) -> (!Maybe Vector2,!IOSt .l)
getWindowPos id ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= (Nothing,ioState)
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID id) windows
	| not found
		= (Nothing,IOStSetDevice (WindowSystemState windows) ioState)
	| otherwise
		# (osdInfo,ioState)		= IOStGetOSDInfo ioState
		  di					= getOSDInfoDocumentInterface osdInfo
		  isSDI					= di==SDI
		  isMDI					= di==MDI
		  (framePtr,clientPtr,getParentPos)
		  						= case (getOSDInfoOSInfo osdInfo) of
		  							Just info -> (info.osFrame,info.osClient,if isMDI (OSgetWindowPos info.osClient) (return (0,0)))
							  		nothing   -> (OSNoWindowPtr, OSNoWindowPtr,return (0,0))
		# (wids,wsH)			= getWindowStateHandleWIDS wsH
		# (tb,ioState)			= getIOToolbox ioState
		# ((wx,wy),tb)			= OSgetWindowPos (if (isSDI && wids.wPtr==clientPtr) framePtr wids.wPtr) tb
		# ((fx,fy),tb)			= getParentPos tb
		# ioState				= setIOToolbox tb ioState
		# ioState				= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
		= (Just {vx=wx-fx,vy=wy-fy},ioState)
-}		
--	Operations that are concerned with the ViewFrame of a window.

moveWindowViewFrame :: Id -> Vector2 -> GUI ps ()
moveWindowViewFrame id v = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return ()
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in 
	        if not found then appIOEnv (ioStSetDevice (WindowSystemState windows))
	        else if getWindowStateHandleWindowKind wsH /= IsWindow
	             then appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
	             else do
			wMetrics <- accIOEnv ioStGetOSWindowMetrics
			wsH <- liftIO (moveWindowViewFrame' wMetrics v wsH)
			appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1))))
	where
		moveWindowViewFrame' :: OSWindowMetrics -> Vector2 -> WindowStateHandle ps -> IO (WindowStateHandle ps)
		moveWindowViewFrame' wMetrics v wsH@(WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH}))) = do
			wH <- CP.moveWindowViewFrame wMetrics v wids wH
			return (WindowStateHandle wids (Just wlsH{wlsHandle=wH}))
		moveWindowViewFrame' _ _ _ = stdWindowFatalError "moveWindowViewFrame" "unexpected window placeholder argument"

getWindowViewFrame :: Id -> GUI ps ViewFrame
getWindowViewFrame id = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return zero
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in
		if not found
		then do
			appIOEnv (ioStSetDevice (WindowSystemState windows))
			return zero
		else do
			wMetrics <- accIOEnv ioStGetOSWindowMetrics
			let viewFrame = getwindowviewframe wMetrics wsH
			appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
			return viewFrame)


-- getwindowviewframe is also used by getWindowOuterSize.
getwindowviewframe :: OSWindowMetrics -> (WindowStateHandle ps) -> ViewFrame
getwindowviewframe wMetrics (WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH})))
	| whKind wH ==IsWindow = rectToRectangle contentRect
	| otherwise = sizeToRectangle (whSize wH)
	where	
		(origin,domainRect,hasHScroll,hasVScroll) = case whWindowInfo wH of
			WindowInfo info	-> (windowOrigin info,windowDomain info,isJust (windowHScroll info),isJust (windowVScroll info))
			_		-> stdWindowFatalError "getWindowViewFrame" "Window has no WindowInfo"
		visScrolls  = osScrollbarsAreVisible wMetrics domainRect (toTuple (whSize wH)) (hasHScroll,hasVScroll)
		contentRect = getWindowContentRect wMetrics visScrolls (posSizeToRect origin (whSize wH))
getwindowviewframe _ _ = stdWindowFatalError "getWindowViewFrame" "unexpected window placeholder argument"

{-
setWindowViewSize :: !Id !Size !(IOSt .l) -> IOSt .l
setWindowViewSize wid reqSize ioState
	# (found,wDevice,ioState)	= IOStGetDevice WindowDevice ioState
	| not found
		= ioState
	# windows					= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)		= getWindowHandlesWindow (toWID wid) windows
	| not found
		= IOStSetDevice (WindowSystemState windows) ioState
	# (wKind,wsH)				= getWindowStateHandleWindowKind wsH
	| wKind<>IsWindow
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	# (wMetrics,ioState)		= IOStGetOSWindowMetrics ioState
	  (diffSize,viewSize,frameSize,wsH)
	  							= validateSize wMetrics reqSize wsH
	| not diffSize
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
	| otherwise
		# (activeWIDS,windows)	= getWindowHandlesActiveWindow windows
		# (osdInfo,ioState)		= IOStGetOSDInfo ioState
		  isSDI					= getOSDInfoDocumentInterface osdInfo==SDI
		  (framePtr,clientPtr,tbHeight)
		  						= case (getOSDInfoOSInfo osdInfo) of
		  							Just info -> (info.osFrame,info.osClient,case info.osToolbar of
		  																		Just {toolbarHeight} -> toolbarHeight
		  																		_					 -> 0
		  										 )
		  							_         -> (OSNoWindowPtr,OSNoWindowPtr,0)
		# (wids,wsH)			= getWindowStateHandleWIDS wsH
		# (tb,ioState)			= getIOToolbox ioState
	//	# tb					= OSsetWindowSize wids.wPtr (toTuple reqSize) True tb
		# tb					= OSsetWindowViewFrameSize wids.wPtr (toTuple viewSize) tb
		# tb					= (if (isSDI && wids.wPtr==clientPtr)
									(OSsetWindowViewFrameSize framePtr (toTuple {viewSize & h=viewSize.h+tbHeight}))
									id
								  ) tb
		# (wsH,tb)				= windowStateSizeAction wMetrics
									(isJust activeWIDS && (fromJust activeWIDS).wId==wid)
									{wsWIDS=wids,wsSize=frameSize,wsUpdateAll=False} wsH tb
		# ioState				= setIOToolbox tb ioState
		= IOStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows)) ioState
where
	validateSize :: !OSWindowMetrics !Size !(WindowStateHandle .pst) -> (!Bool,!Size,!Size,!WindowStateHandle .pst)
	validateSize wMetrics reqSize wsH=:{wshHandle=Just wlsH=:{wlsHandle=wH}}
		# (visHScroll,visVScroll)
							= OSscrollbarsAreVisible wMetrics domainRect (toTuple curSize) (hasHScroll,hasVScroll)
		  newW				= if visVScroll (okSize.w+wMetrics.osmVSliderWidth)  okSize.w	// Correct newW in case of visible vertical   scrollbar
		  newH				= if visHScroll (okSize.h+wMetrics.osmHSliderHeight) okSize.h	// Correct newH in case of visible horizontal scrollbar
		= (okSize<>curSize,okSize,{w=newW,h=newH},wsH)
	where
		(minWidth,minHeight)= OSMinWindowSize
		minSize				= {w=minWidth,h=minHeight}
		maxSize				= maxScrollWindowSize
		okSize				= {w=SetBetween reqSize.w minSize.w maxSize.w,h=SetBetween reqSize.h minSize.h maxSize.h}
		curSize				= wH.whSize
	//	windowInfo			= fromJust wH.whWindowInfo	Mike: fromJust changed into getWindowInfoWindowData
		windowInfo			= getWindowInfoWindowData wH.whWindowInfo
		domainRect			= windowInfo.windowDomain
		hasHScroll			= isJust windowInfo.windowHScroll
		hasVScroll			= isJust windowInfo.windowVScroll
	validateSize _ _ _
		= StdWindowFatalError "setWindowViewSize" "unexpected window placeholder argument"
		
getWindowViewSize :: !Id !(IOSt .l) -> (!Size,!IOSt .l)
getWindowViewSize id ioState
	# (viewFrame,ioState)	= getWindowViewFrame id ioState
	= (rectangleSize viewFrame,ioState)

setWindowOuterSize :: !Id !Size !(IOSt .l) -> IOSt .l
setWindowOuterSize id {w,h} ioState
	# (osdInfo,ioState)	= IOStGetOSDInfo ioState
	  isMDI				= getOSDInfoDocumentInterface osdInfo == MDI
	# ((dw,dh),ioState)	= accIOToolbox (OSstripOuterSize isMDI True) ioState
	= setWindowViewSize id {w=w-dw,h=h-dh} ioState

getWindowOuterSize :: !Id !(IOSt .l) -> (!Size,!IOSt .l)
getWindowOuterSize id ioState
	# (found,wDevice,ioState)		= IOStGetDevice WindowDevice ioState
	| not found
		= (zero,ioState)
	# windows						= WindowSystemStateGetWindowHandles wDevice
	  (found,wsH,windows)			= getWindowHandlesWindow (toWID id) windows
	| not found
		= (zero,IOStSetDevice (WindowSystemState windows) ioState)
	# (wMetrics,ioState)			= IOStGetOSWindowMetrics ioState
	# (viewFrame,wsH)				= getwindowviewframe wMetrics wsH
	# viewFrameSize					= rectangleSize viewFrame
	# (wKind,wsH)					= getWindowStateHandleWindowKind wsH
	| wKind==IsDialog
		# windows					= setWindowHandlesWindow wsH windows
		  outerSize					= exactWindowSize undef undef viewFrameSize undef undef wKind
		= (outerSize,IOStSetDevice (WindowSystemState windows) ioState)
	// Mike
	| wKind==IsGameWindow
		# (wInfo,wsH)				= getWindowStateHandleWindowInfo wsH
		  info						= getWindowInfoGameWindowData wInfo
		  windows					= setWindowHandlesWindow wsH windows
		= (info.gamewindowSize,IOStSetDevice (WindowSystemState windows) ioState)
	//
	| otherwise
		# (wInfo,wsH)				= getWindowStateHandleWindowInfo wsH
		  info						= getWindowInfoWindowData wInfo
		  (hasHScroll,hasVScroll)	= (isJust info.windowHScroll,isJust info.windowVScroll)
		  domain					= RectToRectangle info.windowDomain
		# (osdInfo,ioState)			= IOStGetOSDInfo ioState
		  isMDI						= getOSDInfoDocumentInterface osdInfo == MDI
		# ((dw,dh),ioState)			= accIOToolbox (OSstripOuterSize isMDI True) ioState
		  outerSize					= exactWindowSize wMetrics domain viewFrameSize hasHScroll hasVScroll wKind
		  addOSOuterSize			= {w=outerSize.w+dw,h=outerSize.h+dh}
		  windows					= setWindowHandlesWindow wsH windows
		= (addOSOuterSize,IOStSetDevice (WindowSystemState windows) ioState)
-}

setWindowViewDomain :: Id -> ViewDomain -> GUI ps ()
setWindowViewDomain wId newDomain = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found
	 then return ()
	 else let
	 	windows = windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID wId) windows
	      in
	      	if not found
		then appIOEnv (ioStSetDevice (WindowSystemState windows))
		else if getWindowStateHandleWindowKind wsH /= IsWindow
		     then appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
		     else do
			wMetrics <- accIOEnv ioStGetOSWindowMetrics
			wsH <- liftIO (setwindowviewdomain wMetrics newDomain wsH)
			appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1))))
	where
		setwindowviewdomain :: OSWindowMetrics -> ViewDomain -> WindowStateHandle ps -> IO (WindowStateHandle ps)
		setwindowviewdomain wMetrics newDomain (WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH}))) =
			let
				newDomain1 = validateViewDomain newDomain
				newDomainSize = rectangleSize newDomain
				newDomainRect = rectangleToRect newDomain
				newOrigin = Point2
					{ x = if w1 >= w newDomainSize then rleft newDomainRect else setBetween (x oldOrigin) (rleft newDomainRect) ((rright  newDomainRect)-w1)
					, y = if h1 >= h newDomainSize then rtop  newDomainRect else setBetween (y oldOrigin) (rtop  newDomainRect) ((rbottom newDomainRect)-h1)
					}
				newVisScrolls = osScrollbarsAreVisible wMetrics newDomainRect wSize' hasScrolls
				newContentRect = getWindowContentRect wMetrics newVisScrolls (sizeToRect wSize)
				Rect{rright=w',rbottom=h'} = newContentRect
				osHState = toOSscrollbarRange (rleft newDomainRect,x newOrigin,rright  newDomainRect) w'
				osVState = toOSscrollbarRange (rtop  newDomainRect,y newOrigin,rbottom newDomainRect) h'
				windowInfo1 		= WindowInfo windowInfo{windowDomain=newDomainRect,windowOrigin=newOrigin}
				newViewFrameRect 	= posSizeToRect newOrigin (Size {w=w',h=h'})
				newViewFrame		= rectToRectangle newViewFrameRect
				oldViewFrame		= rectToRectangle oldViewFrameRect
				oldDomainViewMax	= getdomainviewmax oldDomainRect oldViewFrameRect
				newDomainViewMax	= getdomainviewmax newDomainRect newViewFrameRect
				updArea			= if sysLook && oldOrigin==newOrigin && oldDomainViewMax==newDomainViewMax
							  then []
							  else [newViewFrame]
				updState		= UpdateState{oldFrame=oldViewFrame,newFrame=newViewFrame,updArea=updArea}
			  in do
				setwindowslider hasHScroll wMetrics (wPtr wids) True  osHState wSize'
				setwindowslider hasVScroll wMetrics (wPtr wids) False osVState wSize'
				(if null oldItems		-- window has no controls
				 then let 
					wH1 = wH{whWindowInfo=windowInfo1}
				      in
					if null updArea				-- nothing has to updated
					then return (WindowStateHandle wids (Just wlsH{wlsHandle=wH1}))
					else do
						wH2 <- drawWindowLook wMetrics (wPtr wids) (return ()) updState wH1
						return (WindowStateHandle wids (Just wlsH{wlsHandle=wH2}))
				 else let 			-- window has controls
					hMargins = getWindowHMargins   IsWindow wMetrics atts
					vMargins = getWindowVMargins   IsWindow wMetrics atts
					spaces = getWindowItemSpaces IsWindow wMetrics atts
					reqSize	= Size{w=w'-fst hMargins-snd hMargins,h=h'-fst vMargins-snd vMargins}
				      in do
					(_,newItems) <- layoutControls wMetrics hMargins vMargins spaces reqSize minSize [(newDomain,newOrigin)] oldItems
					let wH1 = wH{whWindowInfo=WindowInfo windowInfo,whItems=newItems}
					wH2 <- forceValidWindowClipState wMetrics True (wPtr wids) wH1
					updRgn <- relayoutControls wMetrics (whSelect wH) (whShow wH) newContentRect newContentRect zero zero (wPtr wids) (whDefaultId wH) oldItems (whItems wH)
					wH3 <- drawWindowLook wMetrics (wPtr wids) (return ()) updState wH2
					wH4 <- updateWindowBackgrounds wMetrics updRgn wids wH3
					return (WindowStateHandle wids (Just wlsH{wlsHandle=wH4})))
			where
				atts			= whAtts wH
				wSize			= whSize wH
				wSize'			= toTuple wSize
				(w1,h1)			= wSize'
				windowInfo		= getWindowInfoWindowData (whWindowInfo wH)
				oldDomainRect		= windowDomain windowInfo
				oldOrigin		= windowOrigin windowInfo
				sysLook			= lookSysUpdate (windowLook windowInfo)
				oldItems		= whItems wH
				hasScrolls		= (isJust (windowHScroll windowInfo),isJust (windowVScroll windowInfo))
				(hasHScroll,hasVScroll)	= hasScrolls
				oldVisScrolls		= osScrollbarsAreVisible wMetrics oldDomainRect wSize' hasScrolls
				oldContentRect		= getWindowContentRect wMetrics oldVisScrolls (sizeToRect wSize)
				oldViewFrameRect	= posSizeToRect oldOrigin (rectSize oldContentRect)
				(defMinW,defMinH)	= osMinWindowSize
				minSize			= Size{w=defMinW,h=defMinH}

				setwindowslider :: Bool -> OSWindowMetrics -> OSWindowPtr -> Bool -> (Int,Int,Int,Int) -> (Int,Int) -> IO ()
				setwindowslider hasScroll wMetrics wPtr isHorizontal state maxcoords
					| hasScroll	= osSetWindowSlider wMetrics wPtr isHorizontal state maxcoords
					| otherwise	= return ()

				getdomainviewmax :: Rect -> Rect -> Point2
				getdomainviewmax domainRect viewframeRect =
					Point2 {x=min (rright domainRect) (rright viewframeRect),y=min (rbottom domainRect) (rbottom viewframeRect)}
		setwindowviewdomain _ _ _
			= stdWindowFatalError "setWindowViewDomain" "unexpected window placeholder argument"

getWindowViewDomain :: Id -> GUI ps (Maybe ViewDomain)
getWindowViewDomain id = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return Nothing
	 else let 
	 	windows = windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in
		if not found then do
			appIOEnv (ioStSetDevice (WindowSystemState windows))
			return Nothing
		else if getWindowStateHandleWindowKind wsH /= IsWindow
		     then do
		     		appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
		     		return Nothing
		     else let 
		     		wInfo = getWindowStateHandleWindowInfo wsH	
		  		domain	= rectToRectangle (windowDomain (getWindowInfoWindowData wInfo))
		  	  in do
		  	  	appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH windows1)))
		  	  	return (Just domain))
		  	  	
--	Set and get ScrollFunctions:

setWindowScrollFunction :: Id -> Direction -> ScrollFunction -> GUI ps ()
setWindowScrollFunction wId direction scrollFun = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return ()
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID wId) windows
	      in if not found || getWindowStateHandleWindowKind wsH /= IsWindow then return ()
		 else let wsH1 = setwindowscrollfunction direction scrollFun wsH
		      in appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH1 windows1))))
	where
		setwindowscrollfunction :: Direction -> ScrollFunction -> WindowStateHandle ps -> WindowStateHandle ps
		setwindowscrollfunction direction scrollFun wsH@(WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH})))
			| direction==Horizontal && isJust hScroll =
				let 
					info = windowInfo{windowHScroll=fmap (setScrollFun scrollFun) hScroll}
					wH   = wH{whWindowInfo=WindowInfo info}
				in
					WindowStateHandle wids (Just wlsH{wlsHandle=wH})
			| direction==Vertical && isJust vScroll =
				let
					info = windowInfo{windowVScroll=fmap (setScrollFun scrollFun) vScroll}
					wH   = wH{whWindowInfo=WindowInfo info}
				in
					WindowStateHandle wids (Just wlsH{wlsHandle=wH})
			| otherwise = wsH
			where
				windowInfo	= getWindowInfoWindowData (whWindowInfo wH)
				hScroll		= windowHScroll windowInfo
				vScroll		= windowVScroll windowInfo

				setScrollFun :: ScrollFunction -> ScrollInfo -> ScrollInfo
				setScrollFun f scrollInfo = scrollInfo{scrollFunction=f}
		setwindowscrollfunction _ _ _ =
			stdWindowFatalError "setWindowScrollFunction" "unexpected window placeholder argument"

getWindowScrollFunction :: Id -> Direction -> GUI ps (Maybe ScrollFunction)
getWindowScrollFunction wId direction = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return Nothing
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID wId) windows
	      in
		if not found || getWindowStateHandleWindowKind wsH /= IsWindow then return Nothing
		else return (getwindowscrollfunction direction wsH))
	where
		getwindowscrollfunction :: Direction -> WindowStateHandle ps -> Maybe ScrollFunction
		getwindowscrollfunction direction wsH@(WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH})))
			| direction==Horizontal && isJust hScroll = fmap scrollFunction hScroll
			| direction==Vertical   && isJust vScroll = fmap scrollFunction vScroll
			| otherwise = Nothing
			where
				windowInfo	= getWindowInfoWindowData (whWindowInfo wH)
				hScroll		= windowHScroll windowInfo
				vScroll		= windowVScroll windowInfo
		getwindowscrollfunction _ _ =
			stdWindowFatalError "getWindowScrollFunction" "unexpected window placeholder argument"

--	Operations that are concerned with remaining attributes of windows.

setWindowTitle :: Id -> Title -> GUI ps ()
setWindowTitle id title = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return ()
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in if not found then return ()
		 else do
			osdInfo <- accIOEnv (ioStGetOSDInfo)
		  	let isSDI = getOSDInfoDocumentInterface osdInfo==SDI
		  	let (framePtr,clientPtr) = case (getOSDInfoOSInfo osdInfo) of
				Just info -> (osFrame info, osClient info)
				_         -> (osNoWindowPtr,osNoWindowPtr)
		  	let wids = getWindowStateHandleWIDS wsH
		  	let wsH1 = setWindowStateHandleWindowTitle title wsH
			liftIO (osSetWindowTitle (if isSDI && (wPtr wids) == clientPtr then framePtr else wPtr wids) title)
			appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH1 windows1))))

setWindowCursor :: Id -> CursorShape -> GUI ps ()
setWindowCursor id shape = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return ()
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in if not found
		 then return ()
		 else do
			wsH1 <- liftIO (setwindowcursor shape wsH)
			appIOEnv (ioStSetDevice (WindowSystemState (setWindowHandlesWindow wsH1 windows1))))
	where
		setwindowcursor :: CursorShape -> WindowStateHandle ps -> IO (WindowStateHandle ps)
		setwindowcursor shape (WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH}))) = do
			osSetWindowCursor (wPtr wids) (toCursorCode shape)
			let cursorAtt = WindowCursor shape
			let (replaced,atts) = creplace isWindowCursor cursorAtt (whAtts wH)
			let atts1 = if replaced then atts else (cursorAtt:atts)
			return (WindowStateHandle wids (Just wlsH{wlsHandle=wH{whAtts=atts}}))
		setwindowcursor _ _ = stdWindowFatalError "setWindowCursor" "unexpected window placeholder argument"

getWindowTitle :: Id -> GUI ps (Maybe Title)
getWindowTitle id = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return Nothing
	 else let
	 	windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in if not found
	         then return Nothing
		 else return (Just (getWindowStateHandleWindowTitle wsH)))


getWindowOk :: Id -> GUI ps (Maybe Id)
getWindowOk id = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return Nothing
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in if not found
	         then return Nothing
	     	 else return (getWindowStateHandleDefaultId wsH))
			
getWindowCancel :: Id -> GUI ps (Maybe Id)
getWindowCancel id = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return Nothing
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in if not found
	      	 then return Nothing
	     	 else return (getWindowStateHandleCancelId wsH))

getWindowCursor :: Id -> GUI ps (Maybe CursorShape)
getWindowCursor id = do
	(found,wDevice)	<- accIOEnv (ioStGetDevice WindowDevice)
	(if not found then return Nothing
	 else let
		windows	= windowSystemStateGetWindowHandles wDevice
	  	(found,wsH,windows1) = getWindowHandlesWindow (toWID id) windows
	      in if not found
	         then return Nothing
	      	 else return (getcursor wsH))
	where
		getcursor :: WindowStateHandle ps -> Maybe CursorShape
		getcursor wsH@(WindowStateHandle wids (Just wlsH@(WindowLSHandle {wlsHandle=wH}))) =
			fmap getWindowCursorAtt (find isWindowCursor (whAtts wH))
		getcursor _ = stdWindowFatalError "getWindowCursor" "unexpected window placeholder argument"